import argparse
import logging
import random
import sys
import traceback

try:
    import numpy as np
except ImportError:
    print("ERROR: in order to use slither-simil, you need to install numpy:")
    print("$ pip3 install numpy --user\n")
    sys.exit(-1)

from slither.tools.similarity.encode import load_and_encode, parse_target
from slither.tools.similarity.model import load_model

try:
    from sklearn import decomposition
    import matplotlib.pyplot as plt
except ImportError:
    decomposition = None
    plt = None

logger = logging.getLogger("Slither-simil")


def plot(args: argparse.Namespace) -> None:
    if decomposition is None or plt is None:
        logger.error(
            "ERROR: In order to use plot mode in slither-simil, you need to install sklearn and matplotlib:"
        )
        logger.error("$ pip3 install sklearn matplotlib --user")
        sys.exit(-1)

    try:
        model = args.model
        model = load_model(model)
        # contract = args.contract
        contract, fname = parse_target(args.fname)
        # solc = args.solc
        infile = args.input
        # ext = args.filter
        # nsamples = args.nsamples

        if fname is None or infile is None:
            logger.error("The plot mode requieres fname and input parameters.")
            sys.exit(-1)

        logger.info("Loading data..")
        cache = load_and_encode(infile, **vars(args))

        data = []
        fs = []

        logger.info("Processing data..")
        for (f, c, n), y in cache.items():
            if (c == contract or contract is None) and n == fname:
                fs.append(f)
                data.append(y)

        if len(data) == 0:
            logger.error("No contract was found with function %s", fname)
            sys.exit(-1)

        data = np.array(data)
        pca = decomposition.PCA(n_components=2)
        tdata = pca.fit_transform(data)

        logger.info("Plotting data..")
        plt.figure(figsize=(20, 10))
        assert len(tdata) == len(fs)
        for [x, y], l in zip(tdata, fs, strict=False):
            x = random.gauss(0, 0.01) + x
            y = random.gauss(0, 0.01) + y
            plt.scatter(x, y, c="blue")
            plt.text(x - 0.001, y + 0.001, l)

        logger.info("Saving figure to plot.png..")
        plt.savefig("plot.png", bbox_inches="tight")

    except Exception:
        logger.error(f"Error in {args.filename}")
        logger.error(traceback.format_exc())
        sys.exit(-1)



from slither.core.cfg.node import NodeType, Node
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.core.declarations import Contract
from slither.utils.output import Output
from slither.slithir.operations import InternalCall, OperationWithLValue
from slither.core.variables.state_variable import StateVariable

Result = list[tuple[Node, list[str]]]


def detect_costly_operations_in_loop(contract: Contract) -> Result:
    ret: Result = []
    for f in contract.functions_entry_points:
        if f.is_implemented:
            costly_operations_in_loop(f.entry_point, 0, [], [], ret)

    return ret


def costly_operations_in_loop(
    node: Node | None,
    in_loop_counter: int,
    visited: list[Node],
    calls_stack: list[str],
    ret: Result,
) -> None:
    if node is None:
        return

    if node in visited:
        return
    # shared visited
    visited.append(node)

    if node.type == NodeType.STARTLOOP:
        in_loop_counter += 1
    elif node.type == NodeType.ENDLOOP:
        in_loop_counter -= 1

    if in_loop_counter > 0:
        for ir in node.irs:
            # Ignore Array/Mapping/Struct types for now
            if isinstance(ir, OperationWithLValue) and isinstance(ir.lvalue, StateVariable):
                ret.append((ir.node, calls_stack.copy()))
                break
            if isinstance(ir, (InternalCall)) and ir.function:
                calls_stack.append(node.function.canonical_name)
                costly_operations_in_loop(
                    ir.function.entry_point, in_loop_counter, visited, calls_stack, ret
                )
                calls_stack.pop()

    for son in node.sons:
        costly_operations_in_loop(son, in_loop_counter, visited, calls_stack, ret)


class CostlyOperationsInLoop(AbstractDetector):
    ARGUMENT = "costly-loop"
    HELP = "Costly operations in a loop"
    IMPACT = DetectorClassification.INFORMATIONAL
    # Overall the detector seems precise, but it does not take into account
    # case where there are external calls or internal calls that might read the state
    # variable changes. In these cases the optimization should not be applied
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#costly-operations-inside-a-loop"

    WIKI_TITLE = "Costly operations inside a loop"
    WIKI_DESCRIPTION = (
        "Costly operations inside a loop might waste gas, so optimizations are justified."
    )

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract CostlyOperationsInLoop{

    uint loop_count = 100;
    uint state_variable=0;

    function bad() external{
        for (uint i=0; i < loop_count; i++){
            state_variable++;
        }
    }

    function good() external{
      uint local_variable = state_variable;
      for (uint i=0; i < loop_count; i++){
        local_variable++;
      }
      state_variable = local_variable;
    }
}
```
Incrementing `state_variable` in a loop incurs a lot of gas because of expensive `SSTOREs`, which might lead to an `out-of-gas`."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Use a local variable to hold the loop computation result."

    def _detect(self) -> list[Output]:
        """"""
        results: list[Output] = []
        for c in self.compilation_unit.contracts_derived:
            values = detect_costly_operations_in_loop(c)
            for node, calls_stack in values:
                func = node.function
                info: DETECTOR_INFO = [func, " has costly operations inside a loop:\n"]
                info += ["\t- ", node, "\n"]

                if len(calls_stack) > 0:
                    info.append("\tCalls stack containing the loop:\n")
                    for call in calls_stack:
                        info.extend(["\t\t", call, "\n"])

                res = self.generate_result(info)
                results.append(res)

        return results



import hashlib
import json
import logging
import os
import zipfile
from collections import OrderedDict
from importlib import metadata
from typing import Any, TYPE_CHECKING
from zipfile import ZipFile


from slither.core.cfg.node import Node
from slither.core.declarations import (
    Contract,
    Function,
    Enum,
    Event,
    Structure,
    Pragma,
    FunctionContract,
    CustomError,
)
from slither.core.source_mapping.source_mapping import SourceMapping
from slither.core.variables.local_variable import LocalVariable
from slither.core.variables.variable import Variable
from slither.exceptions import SlitherError
from slither.utils.colors import yellow
from slither.utils.myprettytable import MyPrettyTable

if TYPE_CHECKING:
    from slither.core.compilation_unit import SlitherCompilationUnit
    from slither.detectors.abstract_detector import AbstractDetector

logger = logging.getLogger("Slither")

###################################################################################
###################################################################################
# region Output
###################################################################################
###################################################################################


def output_to_json(filename: str | None, error, results: dict) -> None:
    """

    :param filename: Filename where the json will be written. If None or "-", write to stdout
    :param error: Error to report
    :param results: Results to report
    :param logger: Logger where to log potential info
    :return:
    """
    # Create our encapsulated JSON result.
    json_result = {"success": error is None, "error": error, "results": results}

    if filename == "-":
        filename = None

    # Determine if we should output to stdout
    if filename is None:
        # Write json to console
        print(json.dumps(json_result))
    else:
        # Write json to file
        if os.path.isfile(filename):
            logger.info(yellow(f"{filename} exists already, the overwrite is prevented"))
        else:
            with open(filename, "w", encoding="utf8") as f:
                json.dump(json_result, f, indent=2)


def _output_result_to_sarif(
    detector: dict, detectors_classes: list["AbstractDetector"], sarif: dict
) -> None:
    confidence = "very-high"
    if detector["confidence"] == "Medium":
        confidence = "high"
    elif detector["confidence"] == "Low":
        confidence = "medium"
    elif detector["confidence"] == "Informational":
        confidence = "low"

    risk = "0.0"
    if detector["impact"] == "High":
        risk = "8.0"
    elif detector["impact"] == "Medium":
        risk = "4.0"
    elif detector["impact"] == "Low":
        risk = "3.0"

    detector_class = next(d for d in detectors_classes if detector["check"] == d.ARGUMENT)
    check_id = (
        str(detector_class.IMPACT.value)
        + "-"
        + str(detector_class.CONFIDENCE.value)
        + "-"
        + detector["check"]
    )

    rule = {
        "id": check_id,
        "name": detector["check"],
        "properties": {"precision": confidence, "security-severity": risk},
        "shortDescription": {"text": detector_class.WIKI_TITLE},
        "help": {"text": detector_class.WIKI_RECOMMENDATION},
    }
    # Add the rule if does not exist yet
    if len([x for x in sarif["runs"][0]["tool"]["driver"]["rules"] if x["id"] == check_id]) == 0:
        sarif["runs"][0]["tool"]["driver"]["rules"].append(rule)

    if not detector["elements"]:
        logger.info(yellow("Cannot generate Github security alert for finding without location"))
        logger.info(yellow(detector["description"]))
        logger.info(yellow("This will be supported in a future Slither release"))
        return

    # From 3.19.10 (http://docs.oasis-open.org/sarif/sarif/v2.0/csprd01/sarif-v2.0-csprd01.html)
    # The locations array SHALL NOT contain more than one element unless the condition indicated by the result,
    # if any, can only be corrected by making a change at every location specified in the array.
    finding = detector["elements"][0]
    path = finding["source_mapping"]["filename_relative"]
    start_line = finding["source_mapping"]["lines"][0]
    end_line = finding["source_mapping"]["lines"][-1]

    sarif["runs"][0]["results"].append(
        {
            "ruleId": check_id,
            "message": {"text": detector["description"], "markdown": detector["markdown"]},
            "level": "warning",
            "locations": [
                {
                    "physicalLocation": {
                        "artifactLocation": {"uri": path},
                        "region": {"startLine": start_line, "endLine": end_line},
                    }
                }
            ],
            "partialFingerprints": {"id": detector["id"]},
        }
    )


def output_to_sarif(
    filename: str | None, results: dict, detectors_classes: list[type["AbstractDetector"]]
) -> None:
    """

    :param filename:
    :type filename:
    :param results:
    :type results:
    :return:
    :rtype:
    """

    sarif: dict[str, Any] = {
        "$schema": "https://raw.githubusercontent.com/oasis-tcs/sarif-spec/master/Schemata/sarif-schema-2.1.0.json",
        "version": "2.1.0",
        "runs": [
            {
                "tool": {
                    "driver": {
                        "name": "Slither",
                        "informationUri": "https://github.com/crytic/slither",
                        "version": metadata.version("slither-analyzer"),
                        "rules": [],
                    }
                },
                "results": [],
            }
        ],
    }

    for detector in results.get("detectors", []):
        _output_result_to_sarif(detector, detectors_classes, sarif)

    if filename == "-":
        filename = None

    # Determine if we should output to stdout
    if filename is None:
        # Write json to console
        print(json.dumps(sarif))
    else:
        # Write json to file
        if os.path.isfile(filename):
            logger.info(yellow(f"{filename} exists already, the overwrite is prevented"))
        else:
            with open(filename, "w", encoding="utf8") as f:
                json.dump(sarif, f, indent=2)


# https://docs.python.org/3/library/zipfile.html#zipfile-objects
ZIP_TYPES_ACCEPTED = {
    "lzma": zipfile.ZIP_LZMA,
    "stored": zipfile.ZIP_STORED,
    "deflated": zipfile.ZIP_DEFLATED,
    "bzip2": zipfile.ZIP_BZIP2,
}


def output_to_zip(filename: str, error: str | None, results: dict, zip_type: str = "lzma"):
    """
    Output the results to a zip
    The file in the zip is named slither_results.json
    Note: the json file will not have indentation, as a result the resulting json file will be smaller
    :param zip_type:
    :param filename:
    :param error:
    :param results:
    :return:
    """
    json_result = {"success": error is None, "error": error, "results": results}
    if os.path.isfile(filename):
        logger.info(yellow(f"{filename} exists already, the overwrite is prevented"))
    else:
        with ZipFile(
            filename,
            "w",
            compression=ZIP_TYPES_ACCEPTED.get(zip_type, zipfile.ZIP_LZMA),
        ) as file_desc:
            file_desc.writestr("slither_results.json", json.dumps(json_result).encode("utf8"))


# endregion
###################################################################################
###################################################################################
# region Json generation
###################################################################################
###################################################################################


def _convert_to_description(d: str) -> str:
    if isinstance(d, str):
        return d

    if not isinstance(d, SourceMapping):
        raise SlitherError(f"{d} does not inherit from SourceMapping, conversion impossible")

    if isinstance(d, Node):
        if d.expression:
            return f"{d.expression} ({d.source_mapping})"
        return f"{d!s} ({d.source_mapping})"

    if hasattr(d, "canonical_name"):
        return f"{d.canonical_name} ({d.source_mapping})"

    if hasattr(d, "name"):
        return f"{d.name} ({d.source_mapping})"

    raise SlitherError(f"{type(d)} cannot be converted (no name, or canonical_name")


def _convert_to_markdown(d: str, markdown_root: str) -> str:
    if isinstance(d, str):
        return d

    if not isinstance(d, SourceMapping):
        raise SlitherError(f"{d} does not inherit from SourceMapping, conversion impossible")

    if isinstance(d, Node):
        if d.expression:
            return f"[{d.expression}]({d.source_mapping.to_markdown(markdown_root)})"
        return f"[{d!s}]({d.source_mapping.to_markdown(markdown_root)})"

    if hasattr(d, "canonical_name"):
        return f"[{d.canonical_name}]({d.source_mapping.to_markdown(markdown_root)})"

    if hasattr(d, "name"):
        return f"[{d.name}]({d.source_mapping.to_markdown(markdown_root)})"

    raise SlitherError(f"{type(d)} cannot be converted (no name, or canonical_name")


def _convert_to_id(d: str) -> str:
    """
    Id keeps the source mapping of the node, otherwise we risk to consider two different node as the same
    :param d:
    :return:
    """
    if isinstance(d, str):
        return d

    if not isinstance(d, SourceMapping):
        raise SlitherError(f"{d} does not inherit from SourceMapping, conversion impossible")

    if isinstance(d, Node):
        if d.expression:
            return f"{d.expression} ({d.source_mapping})"
        return f"{d!s} ({d.source_mapping})"

    if isinstance(d, Pragma):
        return f"{d} ({d.source_mapping})"

    if hasattr(d, "canonical_name"):
        return f"{d.canonical_name}"

    if hasattr(d, "name"):
        return f"{d.name}"

    raise SlitherError(f"{type(d)} cannot be converted (no name, or canonical_name")


# endregion
###################################################################################
###################################################################################
# region Internal functions
###################################################################################
###################################################################################


def _create_base_element(
    custom_type: str,
    name: str,
    source_mapping: dict,
    type_specific_fields: dict[
        str,
        dict[
            str,
            str
            | dict[str, int | str | bool | list[int]]
            | dict[str, dict[str, str | dict[str, int | str | bool | list[int]]] | str],
        ]
        | dict[str, str | dict[str, int | str | bool | list[int]]]
        | str
        | list[str],
    ]
    | None = None,
    additional_fields: dict[str, str] | None = None,
) -> dict[str, Any]:
    if additional_fields is None:
        additional_fields = {}
    if type_specific_fields is None:
        type_specific_fields = {}
    element = {"type": custom_type, "name": name, "source_mapping": source_mapping}
    if type_specific_fields:
        element["type_specific_fields"] = type_specific_fields
    if additional_fields:
        element["additional_fields"] = additional_fields
    return element


def _create_parent_element(
    element: SourceMapping,
) -> dict[
    str,
    str
    | dict[str, int | str | bool | list[int]]
    | dict[str, dict[str, str | dict[str, int | str | bool | list[int]]] | str],
]:
    from slither.core.declarations.contract_level import ContractLevel

    if isinstance(element, FunctionContract):
        if element.contract_declarer:
            contract = Output("")
            contract.add_contract(element.contract_declarer)
            return contract.data["elements"][0]
    elif isinstance(element, ContractLevel):
        if element.contract:
            contract = Output("")
            contract.add_contract(element.contract)
            return contract.data["elements"][0]
    elif isinstance(element, (LocalVariable, Node)):
        if element.function:
            function = Output("")
            function.add_function(element.function)
            return function.data["elements"][0]
    return None


SupportedOutput = Variable | Contract | Function | Enum | Event | Structure | Pragma | Node
AllSupportedOutput = str | SupportedOutput


class Output:
    def __init__(
        self,
        info_: str | list[str | SupportedOutput],
        additional_fields: dict | None = None,
        markdown_root: str = "",
        standard_format: bool = True,
    ) -> None:
        if additional_fields is None:
            additional_fields = {}

        # Allow info to be a string to simplify the API
        info: list[str | SupportedOutput]
        if isinstance(info_, str):
            info = [info_]
        else:
            info = info_

        self._data = OrderedDict()
        self._data["elements"] = []
        self._data["description"] = "".join(_convert_to_description(d) for d in info)
        self._data["markdown"] = "".join(_convert_to_markdown(d, markdown_root) for d in info)
        self._data["first_markdown_element"] = ""
        self._markdown_root = markdown_root

        id_txt = "".join(_convert_to_id(d) for d in info)
        self._data["id"] = hashlib.sha3_256(id_txt.encode("utf8")).hexdigest()

        if standard_format:
            to_add = [i for i in info if not isinstance(i, str)]

            for add in to_add:
                self.add(add)

        if additional_fields:
            self._data["additional_fields"] = additional_fields

    def add(self, add: SupportedOutput, additional_fields: dict | None = None) -> None:
        if not self._data["first_markdown_element"]:
            self._data["first_markdown_element"] = add.source_mapping.to_markdown(
                self._markdown_root
            )
        if isinstance(add, Variable):
            self.add_variable(add, additional_fields=additional_fields)
        elif isinstance(add, Contract):
            self.add_contract(add, additional_fields=additional_fields)
        elif isinstance(add, Function):
            self.add_function(add, additional_fields=additional_fields)
        elif isinstance(add, Enum):
            self.add_enum(add, additional_fields=additional_fields)
        elif isinstance(add, Event):
            self.add_event(add, additional_fields=additional_fields)
        elif isinstance(add, Structure):
            self.add_struct(add, additional_fields=additional_fields)
        elif isinstance(add, CustomError):
            self.add_custom_error(add, additional_fields=additional_fields)
        elif isinstance(add, Pragma):
            self.add_pragma(add, additional_fields=additional_fields)
        elif isinstance(add, Node):
            self.add_node(add, additional_fields=additional_fields)
        else:
            raise SlitherError(f"Impossible to add {type(add)} to the json")

    @property
    def data(self) -> dict:
        return self._data

    @property
    def elements(self) -> list[dict]:
        return self._data["elements"]

    # endregion
    ###################################################################################
    ###################################################################################
    # region Variables
    ###################################################################################
    ###################################################################################

    def add_variable(self, variable: Variable, additional_fields: dict | None = None) -> None:
        if additional_fields is None:
            additional_fields = {}
        type_specific_fields = {"parent": _create_parent_element(variable)}
        element = _create_base_element(
            "variable",
            variable.name,
            variable.source_mapping.to_json(),
            type_specific_fields,
            additional_fields,
        )
        self._data["elements"].append(element)

    def add_variables(self, variables: list[Variable]):
        for variable in sorted(variables, key=lambda x: x.name):
            self.add_variable(variable)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Contract
    ###################################################################################
    ###################################################################################

    def add_contract(self, contract: Contract, additional_fields: dict | None = None) -> None:
        if additional_fields is None:
            additional_fields = {}
        element = _create_base_element(
            "contract", contract.name, contract.source_mapping.to_json(), {}, additional_fields
        )
        self._data["elements"].append(element)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Functions
    ###################################################################################
    ###################################################################################

    def add_function(self, function: Function, additional_fields: dict | None = None) -> None:
        if additional_fields is None:
            additional_fields = {}
        type_specific_fields = {
            "parent": _create_parent_element(function),
            "signature": function.full_name,
        }
        element = _create_base_element(
            "function",
            function.name,
            function.source_mapping.to_json(),
            type_specific_fields,
            additional_fields,
        )
        self._data["elements"].append(element)

    def add_functions(self, functions: list[Function], additional_fields: dict | None = None):
        if additional_fields is None:
            additional_fields = {}
        for function in sorted(functions, key=lambda x: x.name):
            self.add_function(function, additional_fields)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Enum
    ###################################################################################
    ###################################################################################

    def add_enum(self, enum: Enum, additional_fields: dict | None = None) -> None:
        if additional_fields is None:
            additional_fields = {}
        type_specific_fields = {"parent": _create_parent_element(enum)}
        element = _create_base_element(
            "enum",
            enum.name,
            enum.source_mapping.to_json(),
            type_specific_fields,
            additional_fields,
        )
        self._data["elements"].append(element)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Structures
    ###################################################################################
    ###################################################################################

    def add_struct(self, struct: Structure, additional_fields: dict | None = None) -> None:
        if additional_fields is None:
            additional_fields = {}
        type_specific_fields = {"parent": _create_parent_element(struct)}
        element = _create_base_element(
            "struct",
            struct.name,
            struct.source_mapping.to_json(),
            type_specific_fields,
            additional_fields,
        )
        self._data["elements"].append(element)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Events
    ###################################################################################
    ###################################################################################

    def add_event(self, event: Event, additional_fields: dict | None = None) -> None:
        if additional_fields is None:
            additional_fields = {}
        type_specific_fields = {
            "parent": _create_parent_element(event),
            "signature": event.full_name,
        }
        element = _create_base_element(
            "event",
            event.name,
            event.source_mapping.to_json(),
            type_specific_fields,
            additional_fields,
        )

        self._data["elements"].append(element)

    # endregion
    ###################################################################################
    ###################################################################################
    # region CustomError
    ###################################################################################
    ###################################################################################

    def add_custom_error(
        self, custom_error: CustomError, additional_fields: dict | None = None
    ) -> None:
        if additional_fields is None:
            additional_fields = {}
        type_specific_fields = {
            "parent": _create_parent_element(custom_error),
            "signature": custom_error.full_name,
        }
        element = _create_base_element(
            "custom_error",
            custom_error.name,
            custom_error.source_mapping.to_json(),
            type_specific_fields,
            additional_fields,
        )

        self._data["elements"].append(element)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Nodes
    ###################################################################################
    ###################################################################################

    def add_node(self, node: Node, additional_fields: dict | None = None) -> None:
        if additional_fields is None:
            additional_fields = {}
        type_specific_fields = {
            "parent": _create_parent_element(node),
        }
        node_name = str(node.expression) if node.expression else ""
        element = _create_base_element(
            "node",
            node_name,
            node.source_mapping.to_json(),
            type_specific_fields,
            additional_fields,
        )
        self._data["elements"].append(element)

    def add_nodes(self, nodes: list[Node]):
        for node in sorted(nodes, key=lambda x: x.node_id):
            self.add_node(node)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Pragma
    ###################################################################################
    ###################################################################################

    def add_pragma(self, pragma: Pragma, additional_fields: dict | None = None) -> None:
        if additional_fields is None:
            additional_fields = {}
        type_specific_fields = {"directive": pragma.directive}
        element = _create_base_element(
            "pragma",
            pragma.version,
            pragma.source_mapping.to_json(),
            type_specific_fields,
            additional_fields,
        )
        self._data["elements"].append(element)

    # endregion
    ###################################################################################
    ###################################################################################
    # region File
    ###################################################################################
    ###################################################################################

    def add_file(self, filename: str, content: str, additional_fields: dict | None = None):
        if additional_fields is None:
            additional_fields = {}
        type_specific_fields = {"filename": filename, "content": content}
        element = _create_base_element("file", type_specific_fields, additional_fields)

        self._data["elements"].append(element)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Pretty Table
    ###################################################################################
    ###################################################################################

    def add_pretty_table(
        self,
        content: MyPrettyTable,
        name: str,
        additional_fields: dict | None = None,
    ):
        if additional_fields is None:
            additional_fields = {}
        type_specific_fields = {"content": content.to_json(), "name": name}
        element = _create_base_element("pretty_table", type_specific_fields, additional_fields)

        self._data["elements"].append(element)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Others
    ###################################################################################
    ###################################################################################

    def add_other(
        self,
        name: str,
        source_mapping: tuple[str, int, int],
        compilation_unit: "SlitherCompilationUnit",
        additional_fields: dict | None = None,
    ) -> None:
        # If this a tuple with (filename, start, end), convert it to a source mapping.
        if additional_fields is None:
            additional_fields = {}
        if isinstance(source_mapping, tuple):
            # Parse the source id
            (filename, start, end) = source_mapping
            source_id = next(
                (
                    source_unit_id
                    for (
                        source_unit_id,
                        source_unit_filename,
                    ) in compilation_unit.source_units.items()
                    if source_unit_filename == filename
                ),
                -1,
            )

            # Convert to a source mapping string
            source_mapping = f"{start}:{end}:{source_id}"

        # If this is a source mapping string, parse it.
        if isinstance(source_mapping, str):
            source_mapping_str = source_mapping
            source_mapping = SourceMapping()
            source_mapping.set_offset(source_mapping_str, compilation_unit)

        # If this is a source mapping object, get the underlying source mapping dictionary
        if isinstance(source_mapping, SourceMapping):
            source_mapping = source_mapping.source_mapping.to_json()

        # Create the underlying element and add it to our resulting json
        element = _create_base_element("other", name, source_mapping, {}, additional_fields)
        self._data["elements"].append(element)



from collections import defaultdict


class Context:
    def __init__(self) -> None:
        super().__init__()
        self._context: dict = {"MEMBERS": defaultdict(None)}

    @property
    def context(self) -> dict:
        """
        Dict used by analysis
        """
        return self._context



from typing import Union, TYPE_CHECKING, Any

from slither.core.solidity_types.type import Type

if TYPE_CHECKING:
    from slither.core.solidity_types.elementary_type import ElementaryType
    from slither.core.solidity_types.type_alias import TypeAliasTopLevel


class MappingType(Type):
    def __init__(
        self,
        type_from: "ElementaryType",
        type_to: Union["MappingType", "TypeAliasTopLevel", "ElementaryType"],
    ) -> None:
        assert isinstance(type_from, Type)
        assert isinstance(type_to, Type)
        super().__init__()
        self._from = type_from
        self._to = type_to

    @property
    def type_from(self) -> Type:
        return self._from

    @property
    def type_to(self) -> Type:
        return self._to

    @property
    def storage_size(self) -> tuple[int, bool]:
        return 32, True

    @property
    def is_dynamic(self) -> bool:
        return True

    def __str__(self) -> str:
        return f"mapping({self._from!s} => {self._to!s})"

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, MappingType):
            return False
        return self.type_from == other.type_from and self.type_to == other.type_to

    def __hash__(self) -> int:
        return hash(str(self))



import sys

try:
    from fasttext import load_model
    from fasttext import train_unsupervised
except ImportError:
    print("ERROR: slither-simil requires fasttext. Install with:")
    print("  pip install slither-analyzer[simil]")
    print("  # or: uv pip install slither-analyzer[simil]")
    print("  # or add fasttext to existing install: pip install fasttext")
    sys.exit(-1)



from slither.core.variables.state_variable import StateVariable
from slither.slithir.variables.variable import SlithIRVariable


class StateIRVariable(StateVariable, SlithIRVariable):
    def __init__(self, state_variable: StateVariable) -> None:
        assert isinstance(state_variable, StateVariable)

        super().__init__()

        # initiate ChildContract
        self.set_contract(state_variable.contract)

        # initiate Variable
        self._name = state_variable.name
        self._initial_expression = state_variable.expression
        self._type = state_variable.type
        self._initialized = state_variable.initialized
        self._visibility = state_variable.visibility
        self._is_constant = state_variable.is_constant

        self._index = 0

        # keep un-ssa version
        if isinstance(state_variable, StateIRVariable):
            self._non_ssa_version = state_variable.non_ssa_version
        else:
            self._non_ssa_version = state_variable

    @property
    def index(self) -> int:
        return self._index

    @index.setter
    def index(self, idx: int) -> None:
        self._index = idx

    @property
    def non_ssa_version(self) -> StateVariable:
        return self._non_ssa_version

    @property
    def ssa_name(self):
        return f"{self._name}_{self.index}"



"""
Module printing summary of the contract
"""

from slither.printers.abstract_printer import AbstractPrinter
from slither.utils.function import get_function_id
from slither.utils.myprettytable import MyPrettyTable


class FunctionIds(AbstractPrinter):
    ARGUMENT = "function-id"
    HELP = "Print the keccak256 signature of the functions"

    WIKI = "https://github.com/trailofbits/slither/wiki/Printer-documentation#function-id"

    def output(self, _filename):
        """
        _filename is not used
        Args:
            _filename(string)
        """

        txt = ""
        all_tables = []
        for contract in self.slither.contracts_derived:
            txt += f"\n{contract.name}:\n"
            table = MyPrettyTable(["Name", "ID"])
            for function in contract.functions:
                if function.is_shadowed or function.is_constructor_variables:
                    continue
                if function.visibility in ["public", "external"]:
                    function_id = get_function_id(function.solidity_signature)
                    table.add_row([function.solidity_signature, f"{function_id:#0{10}x}"])
            for variable in contract.state_variables:
                if variable.visibility in ["public"]:
                    sig = variable.solidity_signature
                    function_id = get_function_id(sig)
                    table.add_row([sig, f"{function_id:#0{10}x}"])
            txt += str(table) + "\n"
            all_tables.append((contract.name, table))

        self.info(txt)

        res = self.generate_output(txt)
        for name, table in all_tables:
            res.add_pretty_table(table, name)

        return res



from enum import Enum

from slither.core.expressions.expression import Expression
from slither.slithir.operations.operation import Operation


class ArgumentType(Enum):
    CALL = 0
    VALUE = 1
    GAS = 2
    DATA = 3


class Argument(Operation):
    def __init__(self, argument: Expression) -> None:
        super().__init__()
        self._argument = argument
        self._type = ArgumentType.CALL
        self._callid: str | None = None

    @property
    def argument(self) -> Expression:
        return self._argument

    @property
    def call_id(self) -> str | None:
        return self._callid

    @call_id.setter
    def call_id(self, c: str) -> None:
        self._callid = c

    @property
    def read(self) -> list[Expression]:
        return [self.argument]

    def set_type(self, t: ArgumentType) -> None:
        assert isinstance(t, ArgumentType)
        self._type = t

    def get_type(self) -> ArgumentType:
        return self._type

    def __str__(self) -> str:
        call_id = "none"
        if self.call_id:
            call_id = f"(id ({self.call_id}))"
        return f"ARG_{self._type.name} {self._argument!s} {call_id}"



"""
Check if an incorrect version of solc is used
"""

import re

from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.formatters.attributes.incorrect_solc import custom_format
from slither.utils.output import Output
from slither.utils.buggy_versions import bugs_by_version

# group:
# 0: ^ > >= < <= (optional)
# 1: ' ' (optional)
# 2: version number
# 3: version number
# 4: version number


PATTERN = re.compile(r"(\^|>|>=|<|<=)?([ ]+)?(\d+)\.(\d+)\.(\d+)")


class IncorrectSolc(AbstractDetector):
    """
    Check if an old version of solc is used
    """

    ARGUMENT = "solc-version"
    HELP = "Incorrect Solidity version"
    IMPACT = DetectorClassification.INFORMATIONAL
    CONFIDENCE = DetectorClassification.HIGH
    LANGUAGE = "solidity"
    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#incorrect-versions-of-solidity"

    WIKI_TITLE = "Incorrect versions of Solidity"

    # region wiki_description
    WIKI_DESCRIPTION = """
`solc` frequently releases new compiler versions. Using an old version prevents access to new Solidity security checks.
We also recommend avoiding complex `pragma` statement."""
    # endregion wiki_description

    # region wiki_recommendation
    WIKI_RECOMMENDATION = """
Deploy with a recent version of Solidity (at least 0.8.0) with no known severe issues.

Use a simple pragma version that allows any of these versions.
Consider using the latest version of Solidity for testing."""
    # endregion wiki_recommendation

    COMPLEX_PRAGMA_TXT = "is too complex"
    OLD_VERSION_TXT = (
        "is an outdated solc version. Use a more recent version (at least 0.8.0), if possible."
    )
    LESS_THAN_TXT = "uses lesser than"

    BUGGY_VERSION_TXT = (
        "contains known severe issues (https://solidity.readthedocs.io/en/latest/bugs.html)"
    )

    # Indicates the allowed versions. Must be formatted in increasing order.
    ALLOWED_VERSIONS = ["0.8.0"]

    def _check_version(self, version: tuple[str, str, str, str, str]) -> str | None:
        op = version[0]
        if op and op not in [">", ">=", "^"]:
            return self.LESS_THAN_TXT
        version_number = ".".join(version[2:])
        if version_number in bugs_by_version and len(bugs_by_version[version_number]):
            bugs = "\n".join([f"\t- {bug}" for bug in bugs_by_version[version_number]])
            return self.BUGGY_VERSION_TXT + f"\n{bugs}"
        return None

    def _check_pragma(self, version: str) -> str | None:
        versions = PATTERN.findall(version)
        if len(versions) == 1:
            version = versions[0]
            return self._check_version(version)
        if len(versions) == 2:
            version_left = versions[0]
            version_right = versions[1]
            # Only allow two elements if the second one is
            # <0.5.0 or <0.6.0
            if version_right not in [
                ("<", "", "0", "5", "0"),
                ("<", "", "0", "6", "0"),
                ("<", "", "0", "7", "0"),
            ]:
                return self.COMPLEX_PRAGMA_TXT
            return self._check_version(version_left)
        return self.COMPLEX_PRAGMA_TXT

    def _detect(self) -> list[Output]:
        """
        Detects pragma statements that allow for outdated solc versions.
        :return: Returns the relevant JSON data for the findings.
        """
        # Detect all version related pragmas and check if they are disallowed.
        results = []
        pragma = self.compilation_unit.pragma_directives
        disallowed_pragmas = {}

        for p in pragma:
            # Skip any pragma directives which do not refer to version
            if len(p.directive) < 1 or p.directive[0] != "solidity":
                continue

            reason = self._check_pragma(p.version)
            if reason is None:
                continue

            if p.version in disallowed_pragmas and reason in disallowed_pragmas[p.version]:
                disallowed_pragmas[p.version][reason].append(p)
            else:
                disallowed_pragmas[p.version] = {reason: [p]}

        # If we found any disallowed pragmas, we output our findings.
        if len(disallowed_pragmas.keys()):
            for p, reasons in disallowed_pragmas.items():
                info: DETECTOR_INFO = []
                for r, vers in reasons.items():
                    info += [f"Version constraint {p} {r}.\nIt is used by:\n"]
                    for ver in vers:
                        info += ["\t- ", ver, "\n"]

                json = self.generate_result(info)

                results.append(json)

        if list(map(int, self.compilation_unit.solc_version.split("."))) < list(
            map(int, self.ALLOWED_VERSIONS[-1].split("."))
        ):
            info = [
                "solc-",
                self.compilation_unit.solc_version,
                " ",
                self.OLD_VERSION_TXT,
                "\n",
            ]

            json = self.generate_result(info)

            # TODO: Once crytic-compile adds config file info, add a source mapping element pointing to
            #       the line in the config that specifies the problematic version of solc

            results.append(json)

        return results

    @staticmethod
    def _format(slither, result):
        custom_format(slither, result)



"""
Module detecting tautologies and contradictions based on types in comparison operations over integers
"""

from slither.core.cfg.node import Node
from slither.core.declarations import Function
from slither.core.declarations.contract import Contract
from slither.core.solidity_types.elementary_type import Int, Uint
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.slithir.operations import Binary, BinaryType
from slither.slithir.variables import Constant
from slither.utils.output import Output


def typeRange(t: str) -> tuple[int, int]:
    bits = int(t.split("int")[1])
    if t in Uint:
        return 0, (2**bits) - 1
    assert t in Int
    v = (2 ** (bits - 1)) - 1
    return -v, v


def _detect_tautology_or_contradiction(low: int, high: int, cval: int, op: BinaryType) -> bool:
    """
    Return true if "[low high] op cval " is always true or always false
    :param low:
    :param high:
    :param cval:
    :param op:
    :return:
    """
    if op == BinaryType.LESS:
        # a < cval
        # its a tautology if
        # high(a) < cval
        # its a contradiction if
        # low(a) >= cval
        return high < cval or low >= cval
    if op == BinaryType.GREATER:
        # a > cval
        # its a tautology if
        # low(a) > cval
        # its a contradiction if
        # high(a) <= cval
        return low > cval or high <= cval
    if op == BinaryType.LESS_EQUAL:
        # a <= cval
        # its a tautology if
        # high(a) <= cval
        # its a contradiction if
        # low(a) > cval
        return (high <= cval) or (low > cval)
    if op == BinaryType.GREATER_EQUAL:
        # a >= cval
        # its a tautology if
        # low(a) >= cval
        # its a contradiction if
        # high(a) < cval
        return (low >= cval) or (high < cval)
    return False


class TypeBasedTautology(AbstractDetector):
    """
    Type-based tautology or contradiction
    """

    ARGUMENT = "tautology"
    HELP = "Tautology or contradiction"
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = (
        "https://github.com/crytic/slither/wiki/Detector-Documentation#tautology-or-contradiction"
    )

    WIKI_TITLE = "Tautology or contradiction"
    WIKI_DESCRIPTION = """Detects expressions that are tautologies or contradictions."""

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract A {
	function f(uint x) public {
		// ...
        if (x >= 0) { // bad -- always true
           // ...
        }
		// ...
	}

	function g(uint8 y) public returns (bool) {
		// ...
        return (y < 512); // bad!
		// ...
	}
}
```
`x` is a `uint256`, so `x >= 0` will be always true.
`y` is a `uint8`, so `y <512` will be always true.
"""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = (
        """Fix the incorrect comparison by changing the value type or the comparison."""
    )

    flip_table = {
        BinaryType.GREATER: BinaryType.LESS,
        BinaryType.GREATER_EQUAL: BinaryType.LESS_EQUAL,
        BinaryType.LESS: BinaryType.GREATER,
        BinaryType.LESS_EQUAL: BinaryType.GREATER_EQUAL,
    }

    def detect_type_based_tautologies(self, contract: Contract) -> list[tuple[Function, set[Node]]]:
        """
        Detects and returns all nodes with tautology/contradiction comparisons (based on type alone).
        :param contract: Contract to detect assignment within.
        :return: A list of nodes with tautolgies/contradictions.
        """

        # Create our result set.
        results: list[tuple[Function, set[Node]]] = []
        allInts = Int + Uint

        # Loop for each function and modifier.
        for function in contract.functions_declared:
            f_results = set()

            for node in function.nodes:
                for ir in node.irs:
                    if isinstance(ir, Binary) and ir.type in self.flip_table:
                        # If neither side is a constant, we can't do much
                        if isinstance(ir.variable_left, Constant):
                            cval = ir.variable_left.value
                            rtype = str(ir.variable_right.type)
                            if rtype in allInts:
                                (low, high) = typeRange(rtype)
                                if _detect_tautology_or_contradiction(
                                    low, high, cval, self.flip_table[ir.type]
                                ):
                                    f_results.add(node)

                        if isinstance(ir.variable_right, Constant):
                            cval = ir.variable_right.value
                            ltype = str(ir.variable_left.type)
                            if ltype in allInts:
                                (low, high) = typeRange(ltype)
                                if _detect_tautology_or_contradiction(low, high, cval, ir.type):
                                    f_results.add(node)
            results.append((function, f_results))

        # Return the resulting set of nodes with tautologies and contradictions
        return results

    def _detect(self) -> list[Output]:
        """
        Detect tautological (or contradictory) comparisons
        """
        results = []
        for contract in self.contracts:
            tautologies = self.detect_type_based_tautologies(contract)
            if tautologies:
                for func, nodes in tautologies:
                    for node in nodes:
                        info = [func, " contains a tautology or contradiction:\n"]
                        info += ["\t- ", node, "\n"]

                        res = self.generate_result(info)
                        results.append(res)

        return results



from slither.solc_parsing.variables.variable_declaration import VariableDeclarationSolc
from slither.core.variables.local_variable_init_from_tuple import LocalVariableInitFromTuple


class LocalVariableInitFromTupleSolc(VariableDeclarationSolc):
    def __init__(
        self, variable: LocalVariableInitFromTuple, variable_data: dict, index: int
    ) -> None:
        super().__init__(variable, variable_data)
        variable.tuple_index = index

    @property
    def underlying_variable(self) -> LocalVariableInitFromTuple:
        # Todo: Not sure how to overcome this with mypy
        assert isinstance(self._variable, LocalVariableInitFromTuple)
        return self._variable

    def _analyze_variable_attributes(self, attributes: dict) -> None:
        """'
        Variable Location
        Can be storage/memory or default
        """
        if "storageLocation" in attributes:
            location = attributes["storageLocation"]
            self.underlying_variable.set_location(location)
        else:
            if "memory" in attributes["type"]:
                self.underlying_variable.set_location("memory")
            elif "storage" in attributes["type"]:
                self.underlying_variable.set_location("storage")
            else:
                self.underlying_variable.set_location("default")

        super()._analyze_variable_attributes(attributes)



import logging
from pathlib import Path

from slither import Slither
from slither.detectors.abstract_detector import AbstractDetector
from slither.detectors.variables.unused_state_variables import UnusedStateVars
from slither.detectors.attributes.incorrect_solc import IncorrectSolc
from slither.detectors.attributes.constant_pragma import ConstantPragma
from slither.detectors.naming_convention.naming_convention import NamingConvention
from slither.detectors.functions.external_function import ExternalFunction
from slither.detectors.variables.could_be_constant import CouldBeConstant
from slither.detectors.variables.could_be_immutable import CouldBeImmutable
from slither.detectors.attributes.const_functions_asm import ConstantFunctionsAsm
from slither.detectors.attributes.const_functions_state import ConstantFunctionsState
from slither.utils.colors import yellow

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("Slither.Format")

all_detectors: dict[str, type[AbstractDetector]] = {
    "unused-state": UnusedStateVars,
    "solc-version": IncorrectSolc,
    "pragma": ConstantPragma,
    "naming-convention": NamingConvention,
    "external-function": ExternalFunction,
    "constable-states": CouldBeConstant,
    "immutable-states": CouldBeImmutable,
    "constant-function-asm": ConstantFunctionsAsm,
    "constant-functions-state": ConstantFunctionsState,
}


def slither_format(slither: Slither, **kwargs: dict) -> None:
    """'
    Keyword Args:
        detectors_to_run (str): Comma-separated list of detectors, defaults to all
    """

    detectors_to_run = choose_detectors(
        kwargs.get("detectors_to_run", "all"), kwargs.get("detectors_to_exclude", "")
    )

    for detector in detectors_to_run:
        slither.register_detector(detector)

    slither.generate_patches = True

    detector_results = slither.run_detectors()
    detector_results = [x for x in detector_results if x]  # remove empty results
    detector_results = [item for sublist in detector_results for item in sublist]  # flatten

    export = Path("crytic-export", "patches")

    export.mkdir(parents=True, exist_ok=True)

    counter_result = 0

    logger.info(yellow("slither-format is in beta, carefully review each patch before merging it."))

    for result in detector_results:
        if "patches" not in result:
            continue
        one_line_description = result["description"].split("\n")[0]

        export_result = Path(export, f"{counter_result}")
        export_result.mkdir(parents=True, exist_ok=True)
        counter_result += 1

        logger.info(f"Issue: {one_line_description}")
        logger.info(f"Generated: ({export_result})")

        for counter, (_, diff) in enumerate(result["patches_diff"].items()):
            filename = f"fix_{counter}.patch"
            path = Path(export_result, filename)
            logger.info(f"\t- {filename}")
            with open(path, "w", encoding="utf8") as f:
                f.write(diff)


# endregion
###################################################################################
###################################################################################
# region Detectors
###################################################################################
###################################################################################


def choose_detectors(
    detectors_to_run: str, detectors_to_exclude: str
) -> list[type[AbstractDetector]]:
    # If detectors are specified, run only these ones
    cls_detectors_to_run: list[type[AbstractDetector]] = []
    exclude = detectors_to_exclude.split(",")
    if detectors_to_run == "all":
        for key, detector in all_detectors.items():
            if key in exclude:
                continue
            cls_detectors_to_run.append(detector)
    else:
        exclude = detectors_to_exclude.split(",")
        for d in detectors_to_run.split(","):
            if d in all_detectors:
                if d in exclude:
                    continue
                cls_detectors_to_run.append(all_detectors[d])
            else:
                raise Exception(f"Error: {d} is not a detector")
    return cls_detectors_to_run


# endregion
###################################################################################
###################################################################################
# region Debug functions
###################################################################################
###################################################################################


def print_patches(number_of_slither_results: int, patches: dict) -> None:
    logger.info("Number of Slither results: " + str(number_of_slither_results))
    number_of_patches = 0
    for file in patches:
        number_of_patches += len(patches[file])
    logger.info("Number of patches: " + str(number_of_patches))
    for file in patches:
        logger.info("Patch file: " + file)
        for patch in patches[file]:
            logger.info("Detector: " + patch["detector"])
            logger.info("Old string: " + patch["old_string"].replace("\n", ""))
            logger.info("New string: " + patch["new_string"].replace("\n", ""))
            logger.info("Location start: " + str(patch["start"]))
            logger.info("Location end: " + str(patch["end"]))


def print_patches_json(number_of_slither_results: int, patches: dict) -> None:
    print("{", end="")
    print('"Number of Slither results":' + '"' + str(number_of_slither_results) + '",')
    print('"Number of patchlets":' + '"' + str(len(patches)) + '"', ",")
    print('"Patchlets":' + "[")
    for index, file in enumerate(patches):
        if index > 0:
            print(",")
        print("{", end="")
        print('"Patch file":' + '"' + file + '",')
        print('"Number of patches":' + '"' + str(len(patches[file])) + '"', ",")
        print('"Patches":' + "[")
        for inner_index, patch in enumerate(patches[file]):
            if inner_index > 0:
                print(",")
            print("{", end="")
            print('"Detector":' + '"' + patch["detector"] + '",')
            print('"Old string":' + '"' + patch["old_string"].replace("\n", "") + '",')
            print('"New string":' + '"' + patch["new_string"].replace("\n", "") + '",')
            print('"Location start":' + '"' + str(patch["start"]) + '",')
            print('"Location end":' + '"' + str(patch["end"]) + '"')
            if "overlaps" in patch:
                print('"Overlaps":' + "Yes")
            print("}", end="")
        print("]", end="")
        print("}", end="")
    print("]", end="")
    print("}")



from .assignment import Assignment
from .binary import Binary, BinaryType
from .call import Call
from .condition import Condition
from .delete import Delete
from .event_call import EventCall
from .high_level_call import HighLevelCall
from .index import Index
from .init_array import InitArray
from .internal_call import InternalCall
from .internal_dynamic_call import InternalDynamicCall
from .library_call import LibraryCall
from .low_level_call import LowLevelCall
from .lvalue import OperationWithLValue
from .member import Member
from .new_array import NewArray
from .new_elementary_type import NewElementaryType
from .new_contract import NewContract
from .new_structure import NewStructure
from .operation import Operation
from .return_operation import Return
from .send import Send
from .solidity_call import SolidityCall
from .transfer import Transfer
from .type_conversion import TypeConversion
from .unary import Unary, UnaryType
from .unpack import Unpack
from .length import Length
from .phi import Phi
from .phi_callback import PhiCallback
from .nop import Nop



import argparse

from slither.tools.kspec_coverage.analysis import run_analysis
from slither import Slither


def kspec_coverage(args: argparse.Namespace) -> None:
    contract = args.contract
    kspec = args.kspec

    slither = Slither(contract, **vars(args))

    compilation_units = slither.compilation_units
    if len(compilation_units) != 1:
        print("Only single compilation unit supported")
        return
    # Run the analysis on the Klab specs
    run_analysis(args, compilation_units[0], kspec)



"""
tests for `slither.core.declarations.Function`.
tests that `tests/test_function.sol` gets translated into correct
`slither.core.declarations.Function` objects or its subclasses
and that these objects behave correctly.
"""

from pathlib import Path

from slither import Slither
from slither.core.declarations.function import FunctionType
from slither.core.solidity_types.elementary_type import ElementaryType
from slither.core.solidity_types.mapping_type import MappingType

TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"
FUNC_DELC_TEST_ROOT = Path(TEST_DATA_DIR, "function_declaration")


def test_functions(solc_binary_path):
    solc_path = solc_binary_path("0.6.12")
    file = Path(FUNC_DELC_TEST_ROOT, "test_function.sol").as_posix()
    slither = Slither(file, solc=solc_path)
    functions = slither.get_contract_from_name("TestFunction")[0].available_functions_as_dict()

    f = functions["external_payable(uint256)"]
    assert f.name == "external_payable"
    assert f.full_name == "external_payable(uint256)"
    assert f.canonical_name == "TestFunction.external_payable(uint256)"
    assert f.solidity_signature == "external_payable(uint256)"
    assert f.signature_str == "external_payable(uint256) returns(uint256)"
    assert f.function_type == FunctionType.NORMAL
    assert f.contains_assembly is False
    assert f.can_reenter() is False
    assert f.can_send_eth() is False
    assert f.is_constructor is False
    assert f.is_fallback is False
    assert f.is_receive is False
    assert f.payable is True
    assert f.visibility == "external"
    assert f.view is False
    assert f.pure is False
    assert f.is_implemented is True
    assert f.is_empty is False
    assert f.parameters[0].name == "_a"
    assert f.parameters[0].type == ElementaryType("uint256")
    assert f.return_type[0] == ElementaryType("uint256")

    f = functions["public_reenter()"]
    assert f.name == "public_reenter"
    assert f.full_name == "public_reenter()"
    assert f.canonical_name == "TestFunction.public_reenter()"
    assert f.solidity_signature == "public_reenter()"
    assert f.signature_str == "public_reenter() returns()"
    assert f.function_type == FunctionType.NORMAL
    assert f.contains_assembly is False
    assert f.can_reenter() is True
    assert f.can_send_eth() is False
    assert f.is_constructor is False
    assert f.is_fallback is False
    assert f.is_receive is False
    assert f.payable is False
    assert f.visibility == "public"
    assert f.view is False
    assert f.pure is False
    assert f.is_implemented is True
    assert f.is_empty is False
    assert f.parameters == []
    assert f.return_type is None

    f = functions["public_payable_reenter_send(bool)"]
    assert f.name == "public_payable_reenter_send"
    assert f.full_name == "public_payable_reenter_send(bool)"
    assert f.canonical_name == "TestFunction.public_payable_reenter_send(bool)"
    assert f.solidity_signature == "public_payable_reenter_send(bool)"
    assert f.signature_str == "public_payable_reenter_send(bool) returns()"
    assert f.function_type == FunctionType.NORMAL
    assert f.contains_assembly is False
    assert f.can_reenter() is True
    assert f.can_send_eth() is True
    assert f.is_constructor is False
    assert f.is_fallback is False
    assert f.is_receive is False
    assert f.payable is True
    assert f.visibility == "public"
    assert f.view is False
    assert f.pure is False
    assert f.is_implemented is True
    assert f.is_empty is False
    assert f.parameters[0].name == "_b"
    assert f.parameters[0].type == ElementaryType("bool")
    assert f.return_type is None

    f = functions["external_send(uint8)"]
    assert f.name == "external_send"
    assert f.full_name == "external_send(uint8)"
    assert f.canonical_name == "TestFunction.external_send(uint8)"
    assert f.solidity_signature == "external_send(uint8)"
    assert f.signature_str == "external_send(uint8) returns()"
    assert f.function_type == FunctionType.NORMAL
    assert f.contains_assembly is False
    assert f.can_reenter() is True
    assert f.can_send_eth() is True
    assert f.is_constructor is False
    assert f.is_fallback is False
    assert f.is_receive is False
    assert f.payable is False
    assert f.visibility == "external"
    assert f.view is False
    assert f.pure is False
    assert f.is_implemented is True
    assert f.is_empty is False
    assert f.parameters[0].name == "_c"
    assert f.parameters[0].type == ElementaryType("uint8")
    assert f.return_type is None

    f = functions["internal_assembly(bytes)"]
    assert f.name == "internal_assembly"
    assert f.full_name == "internal_assembly(bytes)"
    assert f.canonical_name == "TestFunction.internal_assembly(bytes)"
    assert f.solidity_signature == "internal_assembly(bytes)"
    assert f.signature_str == "internal_assembly(bytes) returns(uint256)"
    assert f.function_type == FunctionType.NORMAL
    assert f.contains_assembly is True
    assert f.can_reenter() is False
    assert f.can_send_eth() is False
    assert f.is_constructor is False
    assert f.is_fallback is False
    assert f.is_receive is False
    assert f.payable is False
    assert f.visibility == "internal"
    assert f.view is False
    assert f.pure is False
    assert f.is_implemented is True
    assert f.is_empty is False
    assert f.parameters[0].name == "_d"
    assert f.parameters[0].type == ElementaryType("bytes")
    assert f.return_type[0] == ElementaryType("uint256")

    f = functions["fallback()"]
    assert f.name == "fallback"
    assert f.full_name == "fallback()"
    assert f.canonical_name == "TestFunction.fallback()"
    assert f.solidity_signature == "fallback()"
    assert f.signature_str == "fallback() returns()"
    assert f.function_type == FunctionType.FALLBACK
    assert f.contains_assembly is False
    assert f.can_reenter() is False
    assert f.can_send_eth() is False
    assert f.is_constructor is False
    assert f.is_fallback is True
    assert f.is_receive is False
    assert f.payable is False
    assert f.visibility == "external"
    assert f.view is False
    assert f.pure is False
    assert f.is_implemented is True
    assert f.is_empty is True
    assert f.parameters == []
    assert f.return_type is None

    f = functions["receive()"]
    assert f.name == "receive"
    assert f.full_name == "receive()"
    assert f.canonical_name == "TestFunction.receive()"
    assert f.solidity_signature == "receive()"
    assert f.signature_str == "receive() returns()"
    assert f.function_type == FunctionType.RECEIVE
    assert f.contains_assembly is False
    assert f.can_reenter() is False
    assert f.can_send_eth() is False
    assert f.is_constructor is False
    assert f.is_fallback is False
    assert f.is_receive is True
    assert f.payable is True
    assert f.visibility == "external"
    assert f.view is False
    assert f.pure is False
    assert f.is_implemented is True
    assert f.is_empty is True
    assert f.parameters == []
    assert f.return_type is None

    f = functions["constructor(address)"]
    assert f.name == "constructor"
    assert f.full_name == "constructor(address)"
    assert f.canonical_name == "TestFunction.constructor(address)"
    assert f.solidity_signature == "constructor(address)"
    assert f.signature_str == "constructor(address) returns()"
    assert f.function_type == FunctionType.CONSTRUCTOR
    assert f.contains_assembly is False
    assert f.can_reenter() is False
    assert f.can_send_eth() is False
    assert f.is_constructor
    assert f.is_fallback is False
    assert f.is_receive is False
    assert f.payable is True
    assert f.visibility == "public"
    assert f.view is False
    assert f.pure is False
    assert f.is_implemented is True
    assert f.is_empty is True
    assert f.parameters[0].name == "_e"
    assert f.parameters[0].type == ElementaryType("address")
    assert f.return_type is None

    f = functions["private_view()"]
    assert f.name == "private_view"
    assert f.full_name == "private_view()"
    assert f.canonical_name == "TestFunction.private_view()"
    assert f.solidity_signature == "private_view()"
    assert f.signature_str == "private_view() returns(bool)"
    assert f.function_type == FunctionType.NORMAL
    assert f.contains_assembly is False
    assert f.can_reenter() is False
    assert f.can_send_eth() is False
    assert f.is_constructor is False
    assert f.is_fallback is False
    assert f.is_receive is False
    assert f.payable is False
    assert f.visibility == "private"
    assert f.view is True
    assert f.pure is False
    assert f.is_implemented is True
    assert f.is_empty is False
    assert f.parameters == []
    assert f.return_type[0] == ElementaryType("bool")

    f = functions["public_pure()"]
    assert f.name == "public_pure"
    assert f.full_name == "public_pure()"
    assert f.canonical_name == "TestFunction.public_pure()"
    assert f.solidity_signature == "public_pure()"
    assert f.signature_str == "public_pure() returns(bool)"
    assert f.function_type == FunctionType.NORMAL
    assert f.contains_assembly is False
    assert f.can_reenter() is False
    assert f.can_send_eth() is False
    assert f.is_constructor is False
    assert f.is_fallback is False
    assert f.is_receive is False
    assert f.payable is False
    assert f.visibility == "public"
    assert f.view is True
    assert f.pure is True
    assert f.is_implemented is True
    assert f.is_empty is False
    assert f.parameters == []
    assert f.return_type[0] == ElementaryType("bool")

    f = functions["q()"]
    assert f.name == "q"
    assert f.full_name == "q()"
    assert f.canonical_name == "TestFunction.q()"
    assert f.solidity_signature == "q()"
    assert f.signature_str == "q() returns()"
    assert f.function_type == FunctionType.NORMAL
    assert f.contains_assembly is False
    assert f.can_reenter() is False
    assert f.can_send_eth() is False
    assert f.is_constructor is False
    assert f.is_fallback is False
    assert f.is_receive is False
    assert f.payable is False
    assert f.visibility == "public"
    assert f.view is True
    assert f.pure is True
    assert f.is_implemented is True
    assert f.is_empty is True
    assert f.parameters == []
    assert f.return_type is None
    assert len(f.modifiers) == 1 and f.modifiers[0].name == "m"


def test_function_can_send_eth(solc_binary_path):
    solc_path = solc_binary_path("0.6.12")
    file = Path(FUNC_DELC_TEST_ROOT, "test_function.sol").as_posix()
    slither = Slither(file, solc=solc_path)
    compilation_unit = slither.compilation_units[0]
    functions = compilation_unit.get_contract_from_name("TestFunctionCanSendEth")[
        0
    ].available_functions_as_dict()

    assert functions["send_direct()"].can_send_eth() is True
    assert functions["transfer_direct()"].can_send_eth() is True
    assert functions["call_direct()"].can_send_eth() is True
    assert functions["highlevel_call_direct()"].can_send_eth() is True

    assert functions["send_via_internal()"].can_send_eth() is True
    assert functions["transfer_via_internal()"].can_send_eth() is True
    assert functions["call_via_internal()"].can_send_eth() is True
    assert functions["highlevel_call_via_internal()"].can_send_eth() is True

    assert functions["send_via_external()"].can_send_eth() is False
    assert functions["transfer_via_external()"].can_send_eth() is False
    assert functions["call_via_external()"].can_send_eth() is False
    assert functions["highlevel_call_via_external()"].can_send_eth() is False


def test_reentrant(solc_binary_path):
    solc_path = solc_binary_path("0.8.10")
    file = Path(FUNC_DELC_TEST_ROOT, "test_function_reentrant.sol").as_posix()
    slither = Slither(file, solc=solc_path)
    compilation_unit = slither.compilation_units[0]
    functions = compilation_unit.get_contract_from_name("TestReentrant")[
        0
    ].available_functions_as_dict()

    assert functions["is_reentrant()"].is_reentrant
    assert not functions["is_non_reentrant()"].is_reentrant
    assert not functions["internal_and_not_reentrant()"].is_reentrant
    assert not functions["internal_and_not_reentrant2()"].is_reentrant
    assert functions["internal_and_could_be_reentrant()"].is_reentrant
    assert functions["internal_and_reentrant()"].is_reentrant


def test_public_variable(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.6.12")
    file = Path(FUNC_DELC_TEST_ROOT, "test_function.sol").as_posix()
    slither = Slither(file, solc=solc_path)
    contracts = slither.get_contract_from_name("TestFunction")
    assert len(contracts) == 1
    contract = contracts[0]
    var = contract.get_state_variable_from_name("info")
    assert var
    assert var.solidity_signature == "info()"
    assert var.signature_str == "info() returns(bytes32)"
    assert var.visibility == "public"
    assert var.type == ElementaryType("bytes32")


def test_vyper_functions(slither_from_vyper_source) -> None:
    with slither_from_vyper_source(
        """
balances: public(HashMap[address, uint256])
allowances: HashMap[address, HashMap[address, uint256]]
@pure
@internal
def add(x: int128, y: int128) -> int128:
    return x + y
@external
def __init__():
    pass
@external
def withdraw():
    raw_call(msg.sender, b"", value= self.balances[msg.sender])
@external
@nonreentrant("lock")
def withdraw_locked():
    self.withdraw_locked_internal()
@internal
def withdraw_locked_internal():
    raw_call(msg.sender, b"", value= self.balances[msg.sender])
@payable
@external
def __default__():
    pass
    """
    ) as sl:
        contract = sl.contracts[0]
        functions = contract.available_functions_as_dict()

        f = functions["add(int128,int128)"]
        assert f.function_type == FunctionType.NORMAL
        assert f.visibility == "internal"
        assert not f.payable
        assert f.view is False
        assert f.pure is True
        assert f.parameters[0].name == "x"
        assert f.parameters[0].type == ElementaryType("int128")
        assert f.parameters[1].name == "y"
        assert f.parameters[1].type == ElementaryType("int128")
        assert f.return_type[0] == ElementaryType("int128")

        f = functions["__init__()"]
        assert f.function_type == FunctionType.CONSTRUCTOR
        assert f.visibility == "external"
        assert not f.payable
        assert not f.view
        assert not f.pure
        assert not f.is_implemented
        assert f.is_empty

        f = functions["__default__()"]
        assert f.function_type == FunctionType.FALLBACK
        assert f.visibility == "external"
        assert f.payable
        assert not f.view
        assert not f.pure
        assert not f.is_implemented
        assert f.is_empty

        f = functions["withdraw()"]
        assert f.function_type == FunctionType.NORMAL
        assert f.visibility == "external"
        assert not f.payable
        assert not f.view
        assert not f.pure
        assert f.can_send_eth()
        assert f.can_reenter()
        assert f.is_implemented
        assert not f.is_empty

        f = functions["withdraw_locked()"]
        assert f.is_reentrant is False
        assert f.is_implemented
        assert not f.is_empty

        f = functions["withdraw_locked_internal()"]
        assert f.is_reentrant is False
        assert f.visibility == "internal"

        var = contract.get_state_variable_from_name("balances")
        assert var
        assert var.solidity_signature == "balances(address)"
        assert var.signature_str == "balances(address) returns(uint256)"
        assert var.visibility == "public"
        assert var.type == MappingType(ElementaryType("address"), ElementaryType("uint256"))

        var = contract.get_state_variable_from_name("allowances")
        assert var
        assert var.solidity_signature == "allowances(address,address)"
        assert var.signature_str == "allowances(address,address) returns(uint256)"
        assert var.visibility == "internal"
        assert var.type == MappingType(
            ElementaryType("address"),
            MappingType(ElementaryType("address"), ElementaryType("uint256")),
        )



from fractions import Fraction

from slither.exceptions import SlitherError


def convert_string_to_fraction(val: str | bytes | int) -> Fraction:
    if isinstance(val, bytes):
        return int.from_bytes(val, byteorder="big")
    if isinstance(val, int):
        return Fraction(val)
    if val.startswith(("0x", "0X")):
        return Fraction(int(val, 16))

    # Fractions do not support underscore separators (on Python <3.11)
    val = val.replace("_", "")

    if "e" in val or "E" in val:
        base, expo = val.split("e") if "e" in val else val.split("E")
        base, expo = Fraction(base), int(expo)
        # The resulting number must be < 2**256-1, otherwise solc
        # Would not be able to compile it
        # 10**77 is the largest exponent that fits
        # See https://github.com/ethereum/solidity/blob/9e61f92bd4d19b430cb8cb26f1c7cf79f1dff380/libsolidity/ast/Types.cpp#L1281-L1290
        if expo > 77:
            if base != Fraction(0):
                raise SlitherError(
                    f"{base}e{expo} is too large to fit in any Solidity integer size"
                )
            return 0
        return Fraction(base) * Fraction(10**expo)

    return Fraction(val)


def convert_string_to_int(val: str | int) -> int:
    return int(convert_string_to_fraction(val))



from pathlib import Path
from slither import Slither

TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"


def test_overrides(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.15")
    slither = Slither(Path(TEST_DATA_DIR, "virtual_overrides.sol").as_posix(), solc=solc_path)

    test = slither.get_contract_from_name("Test")[0]
    test_virtual_func = test.get_function_from_full_name("myVirtualFunction()")
    assert test_virtual_func.is_virtual
    assert not test_virtual_func.is_override
    x = test.get_functions_overridden_by(test_virtual_func)
    assert len(x) == 0
    x = test_virtual_func.overridden_by
    assert len(x) == 5
    assert set(i.canonical_name for i in x) == {
        "A.myVirtualFunction()",
        "C.myVirtualFunction()",
        "X.myVirtualFunction()",
    }

    a = slither.get_contract_from_name("A")[0]
    a_virtual_func = a.get_function_from_full_name("myVirtualFunction()")
    assert a_virtual_func.is_virtual
    assert a_virtual_func.is_override
    x = a.get_functions_overridden_by(a_virtual_func)
    assert len(x) == 2
    assert set(i.canonical_name for i in x) == {"Test.myVirtualFunction()"}

    b = slither.get_contract_from_name("B")[0]
    b_virtual_func = b.get_function_from_full_name("myVirtualFunction()")
    assert not b_virtual_func.is_virtual
    assert b_virtual_func.is_override
    x = b.get_functions_overridden_by(b_virtual_func)
    assert len(x) == 2
    assert set(i.canonical_name for i in x) == {"A.myVirtualFunction()"}
    assert len(b_virtual_func.overridden_by) == 0

    c = slither.get_contract_from_name("C")[0]
    c_virtual_func = c.get_function_from_full_name("myVirtualFunction()")
    assert not c_virtual_func.is_virtual
    assert c_virtual_func.is_override
    x = c.get_functions_overridden_by(c_virtual_func)
    assert len(x) == 2
    # C should not override B as they are distinct leaves in the inheritance tree
    assert set(i.canonical_name for i in x) == {"Test.myVirtualFunction()"}

    y = slither.get_contract_from_name("Y")[0]
    y_virtual_func = y.get_function_from_full_name("myVirtualFunction()")
    assert y_virtual_func.is_virtual
    assert not y_virtual_func.is_override
    x = y_virtual_func.overridden_by
    assert len(x) == 1
    assert x[0].canonical_name == "Z.myVirtualFunction()"

    z = slither.get_contract_from_name("Z")[0]
    z_virtual_func = z.get_function_from_full_name("myVirtualFunction()")
    assert z_virtual_func.is_virtual
    assert z_virtual_func.is_override
    x = z.get_functions_overridden_by(z_virtual_func)
    assert len(x) == 4
    assert set(i.canonical_name for i in x) == {"Y.myVirtualFunction()", "X.myVirtualFunction()"}

    k = slither.get_contract_from_name("K")[0]
    k_virtual_func = k.get_function_from_full_name("a()")
    assert not k_virtual_func.is_virtual
    assert k_virtual_func.is_override
    assert len(k_virtual_func.overrides) == 3
    x = k_virtual_func.overrides
    assert set(i.canonical_name for i in x) == {"I.a()"}

    i = slither.get_contract_from_name("I")[0]
    i_virtual_func = i.get_function_from_full_name("a()")
    assert i_virtual_func.is_virtual
    assert not i_virtual_func.is_override
    assert len(i_virtual_func.overrides) == 0
    x = i_virtual_func.overridden_by
    assert len(x) == 1
    assert x[0].canonical_name == "K.a()"


def test_virtual_override_references_and_implementations(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.15")
    file = Path(TEST_DATA_DIR, "virtual_overrides.sol").as_posix()
    slither = Slither(file, solc=solc_path)
    funcs = slither.offset_to_objects(file, 29)
    assert len(funcs) == 1
    func = funcs.pop()
    assert func.canonical_name == "Test.myVirtualFunction()"
    assert {(x.start, x.end) for x in slither.offset_to_implementations(file, 29)} == {
        (20, 73),
        (102, 164),
        (274, 328),
        (357, 419),
    }

    funcs = slither.offset_to_objects(file, 111)
    assert len(funcs) == 1
    func = funcs.pop()
    assert func.canonical_name == "A.myVirtualFunction()"
    # A.myVirtualFunction() is implemented in A and also overridden in B
    assert {(x.start, x.end) for x in slither.offset_to_implementations(file, 111)} == {
        (102, 164),
        (190, 244),
    }

    # X is inherited by Z and Z.myVirtualFunction() overrides X.myVirtualFunction()
    assert {(x.start, x.end) for x in slither.offset_to_references(file, 341)} == {
        (514, 515),
        (570, 571),
    }
    # The reference to X in inheritance specifier is the definition of Z
    assert {(x.start, x.end) for x in slither.offset_to_definitions(file, 514)} == {(341, 343)}
    # The reference to X in the function override specifier is the definition of Z
    assert {(x.start, x.end) for x in slither.offset_to_definitions(file, 570)} == {(341, 343)}

    # Y is inherited by Z and Z.myVirtualFunction() overrides Y.myVirtualFunction()
    assert {(x.start, x.end) for x in slither.offset_to_references(file, 432)} == {
        (511, 512),
        (567, 568),
    }
    # The reference to Y in inheritance specifier is the definition of Z
    assert {(x.start, x.end) for x in slither.offset_to_definitions(file, 511)} == {(432, 434)}
    # The reference to Y in the function override specifier is the definition of Z
    assert {(x.start, x.end) for x in slither.offset_to_definitions(file, 567)} == {(432, 434)}

    # Name is abstract and has no implementation. It is inherited and implemented by Name2
    assert {(x.start, x.end) for x in slither.offset_to_implementations(file, 612)} == {(657, 718)}


def test_virtual_is_implemented(solc_binary_path):
    solc_path = solc_binary_path("0.8.15")
    file = Path(TEST_DATA_DIR, "virtual_overrides.sol").as_posix()
    slither = Slither(file, solc=solc_path)

    test2 = slither.get_contract_from_name("Test2")[0]
    f = test2.get_function_from_full_name("f()")
    assert f.is_virtual
    assert not f.is_implemented

    a2 = slither.get_contract_from_name("A2")[0]
    f = a2.get_function_from_full_name("f()")
    assert f.is_virtual
    assert f.is_implemented

    # Test.2f() is not implemented, but A2 inherits from Test2 and overrides f()
    assert {(x.start, x.end) for x in slither.offset_to_implementations(file, 759)} == {(809, 853)}



import re

from slither.core.compilation_unit import SlitherCompilationUnit
from slither.formatters.utils.patches import create_patch


def custom_format(compilation_unit: SlitherCompilationUnit, result):
    for file_scope in compilation_unit.scopes.values():
        elements = result["elements"]
        for element in elements:
            target_contract = file_scope.get_contract_from_name(
                element["type_specific_fields"]["parent"]["name"]
            )
            if target_contract:
                function = target_contract.get_function_from_full_name(
                    element["type_specific_fields"]["signature"]
                )
                if function:
                    _patch(
                        file_scope,
                        result,
                        element["source_mapping"]["filename_absolute"],
                        int(function.parameters_src().source_mapping.start),
                        int(function.returns_src().source_mapping.start),
                    )


def _patch(
    compilation_unit: SlitherCompilationUnit, result, in_file, modify_loc_start, modify_loc_end
):
    in_file_str = compilation_unit.core.source_code[in_file].encode("utf8")
    old_str_of_interest = in_file_str[modify_loc_start:modify_loc_end]
    # Search for 'public' keyword which is in-between the function name and modifier name (if present)
    # regex: 'public' could have spaces around or be at the end of the line
    m = re.search(r"((\spublic)\s+)|(\spublic)$|(\)public)$", old_str_of_interest.decode("utf8"))
    if m is None:
        # No visibility specifier exists; public by default.
        create_patch(
            result,
            in_file,
            # start after the function definition's closing paranthesis
            modify_loc_start + len(old_str_of_interest.decode("utf8").split(")")[0]) + 1,
            # end is same as start because we insert the keyword `external` at that location
            modify_loc_start + len(old_str_of_interest.decode("utf8").split(")")[0]) + 1,
            "",
            " external",
        )  # replace_text is `external`
    else:
        create_patch(
            result,
            in_file,
            # start at the keyword `public`
            modify_loc_start + m.span()[0] + 1,
            # end after the keyword `public` = start + len('public'')
            modify_loc_start + m.span()[0] + 1 + len("public"),
            "public",
            "external",
        )



from collections import defaultdict


def create_patch_with_line(
    result: dict,
    file: str,
    start: int,
    end: int,
    old_str: str | bytes,
    new_str: str | bytes,
    line_no: int,
) -> None:
    if isinstance(old_str, bytes):
        old_str = old_str.decode("utf8")
    if isinstance(new_str, bytes):
        new_str = new_str.decode("utf8")
    p = {
        "start": start,
        "end": end,
        "old_string": old_str,
        "new_string": new_str,
        "line_number": line_no,
    }
    if "patches" not in result:
        result["patches"] = defaultdict(list)
    if p not in result["patches"][file]:
        result["patches"][file].append(p)



import logging
import re
import uuid
from collections import namedtuple
from enum import Enum as PythonEnum
from pathlib import Path
from collections.abc import Sequence

from slither.core.compilation_unit import SlitherCompilationUnit
from slither.core.declarations import SolidityFunction, EnumContract, StructureContract
from slither.core.declarations.contract import Contract
from slither.core.declarations.function_top_level import FunctionTopLevel
from slither.core.declarations.top_level import TopLevel
from slither.core.declarations.solidity_variables import SolidityCustomRevert
from slither.core.solidity_types import MappingType, ArrayType
from slither.core.solidity_types.type import Type
from slither.core.solidity_types.user_defined_type import UserDefinedType
from slither.exceptions import SlitherException
from slither.slithir.operations import NewContract, TypeConversion, SolidityCall, InternalCall
from slither.tools.flattening.export.export import (
    Export,
    export_as_json,
    save_to_zip,
    save_to_disk,
)

logger = logging.getLogger("Slither-flat")
logger.setLevel(logging.INFO)

# index: where to start
# patch_type:
#   - public_to_external: public to external (external-to-public)
#   - calldata_to_memory: calldata to memory (external-to-public)
#   - line_removal: remove the line (remove-assert)
Patch = namedtuple("PatchExternal", ["index", "patch_type"])


class Strategy(PythonEnum):
    MostDerived = 0
    OneFile = 1
    LocalImport = 2


STRATEGIES_NAMES = ",".join([i.name for i in Strategy])

DEFAULT_EXPORT_PATH = Path("crytic-export/flattening")


class Flattening:
    def __init__(
        self,
        compilation_unit: SlitherCompilationUnit,
        external_to_public=False,
        remove_assert=False,
        convert_library_to_internal=False,
        private_to_internal=False,
        export_path: str | None = None,
        pragma_solidity: str | None = None,
    ):
        self._source_codes: dict[Contract, str] = {}
        self._source_codes_top_level: dict[TopLevel, str] = {}
        self._compilation_unit: SlitherCompilationUnit = compilation_unit
        self._external_to_public = external_to_public
        self._remove_assert = remove_assert
        self._use_abi_encoder_v2 = False
        self._convert_library_to_internal = convert_library_to_internal
        self._private_to_internal = private_to_internal
        self._pragma_solidity = pragma_solidity

        self._export_path: Path = DEFAULT_EXPORT_PATH if export_path is None else Path(export_path)

        self._check_abi_encoder_v2()

        for contract in compilation_unit.contracts:
            self._get_source_code(contract)

        self._get_source_code_top_level(compilation_unit.structures_top_level)
        self._get_source_code_top_level(compilation_unit.enums_top_level)
        self._get_source_code_top_level(compilation_unit.events_top_level)
        self._get_source_code_top_level(compilation_unit.custom_errors)
        self._get_source_code_top_level(compilation_unit.variables_top_level)
        self._get_source_code_top_level(compilation_unit.functions_top_level)

    def _get_source_code_top_level(self, elems: Sequence[TopLevel]) -> None:
        for elem in elems:
            self._source_codes_top_level[elem] = elem.source_mapping.content

    def _check_abi_encoder_v2(self):
        """
        Check if ABIEncoderV2 is required
        Set _use_abi_encorder_v2
        :return:
        """
        for p in self._compilation_unit.pragma_directives:
            if "ABIEncoderV2" in str(p.directive):
                self._use_abi_encoder_v2 = True
                return

    def _get_source_code(self, contract: Contract):
        """
        Save the source code of the contract in self._source_codes
        Patch the source code
        :param contract:
        :return:
        """
        src_mapping = contract.source_mapping
        src_bytes = self._compilation_unit.core.source_code[src_mapping.filename.absolute].encode(
            "utf8"
        )

        to_patch = []
        # interface must use external
        if self._external_to_public and not contract.is_interface:
            for f in contract.functions_declared:
                # fallback must be external
                if f.is_fallback or f.is_constructor_variables:
                    continue
                if f.visibility == "external":
                    attributes_start = (
                        f.parameters_src().source_mapping.start
                        + f.parameters_src().source_mapping.length
                    )
                    attributes_end = f.returns_src().source_mapping.start
                    attributes = src_bytes[attributes_start:attributes_end].decode("utf8")
                    regex = re.search(r"((\sexternal)\s+)|(\sexternal)$|(\)external)$", attributes)
                    if regex:
                        to_patch.append(
                            Patch(
                                attributes_start + regex.span()[0] + 1,
                                "public_to_external",
                            )
                        )
                    else:
                        raise SlitherException(f"External keyword not found {f.name} {attributes}")

                    for var in f.parameters:
                        if var.location == "calldata":
                            calldata_start = var.source_mapping.start
                            calldata_end = calldata_start + var.source_mapping.length
                            calldata_idx = src_bytes[calldata_start:calldata_end].find(" calldata ")
                            to_patch.append(
                                Patch(
                                    calldata_start + calldata_idx + 1,
                                    "calldata_to_memory",
                                )
                            )

        if self._convert_library_to_internal and contract.is_library:
            for f in contract.functions_declared:
                visibility = ""
                if f.visibility in ["external", "public"]:
                    visibility = f.visibility
                    attributes_start = (
                        f.parameters_src().source_mapping["start"]
                        + f.parameters_src().source_mapping["length"]
                    )
                    attributes_end = f.returns_src().source_mapping["start"]
                    attributes = src_bytes[attributes_start:attributes_end].decode("utf8")
                    regex = (
                        re.search(r"((\sexternal)\s+)|(\sexternal)$|(\)external)$", attributes)
                        if visibility == "external"
                        else re.search(r"((\spublic)\s+)|(\spublic)$|(\)public)$", attributes)
                    )
                    if regex:
                        to_patch.append(
                            Patch(
                                attributes_start + regex.span()[0] + 1,
                                "external_to_internal"
                                if visibility == "external"
                                else "public_to_internal",
                            )
                        )
                    else:
                        raise SlitherException(
                            f"{visibility} keyword not found {f.name} {attributes}"
                        )

        if self._private_to_internal:
            for variable in contract.state_variables_declared:
                if variable.visibility == "private":
                    attributes_start = variable.source_mapping.start
                    attributes_end = attributes_start + variable.source_mapping.length
                    attributes = src_bytes[attributes_start:attributes_end].decode("utf8")
                    regex = re.search(r" private ", attributes)
                    if regex:
                        to_patch.append(
                            Patch(
                                attributes_start + regex.span()[0] + 1,
                                "private_to_internal",
                            )
                        )
                    else:
                        raise SlitherException(
                            f"private keyword not found {variable.name} {attributes}"
                        )

        if self._remove_assert:
            for function in contract.functions_and_modifiers_declared:
                for node in function.nodes:
                    for ir in node.irs:
                        if isinstance(ir, SolidityCall) and ir.function == SolidityFunction(
                            "assert(bool)"
                        ):
                            to_patch.append(Patch(node.source_mapping.start, "line_removal"))
                            logger.info(
                                f"Code commented: {node.expression} ({node.source_mapping})"
                            )

        to_patch.sort(key=lambda x: x.index, reverse=True)

        content = src_mapping.content.encode("utf8")
        start = src_mapping.start
        for patch in to_patch:
            patch_type = patch.patch_type
            index = patch.index
            index = index - start
            if patch_type == "public_to_external":
                content = (
                    content[:index].decode("utf8")
                    + "public"
                    + content[index + len("external") :].decode("utf8")
                )
            elif patch_type == "external_to_internal":
                content = (
                    content[:index].decode("utf8")
                    + "internal"
                    + content[index + len("external") :].decode("utf8")
                )
            elif patch_type == "public_to_internal":
                content = (
                    content[:index].decode("utf8")
                    + "internal"
                    + content[index + len("public") :].decode("utf8")
                )
            elif patch_type == "private_to_internal":
                content = (
                    content[:index].decode("utf8")
                    + "internal"
                    + content[index + len("private") :].decode("utf8")
                )
            elif patch_type == "calldata_to_memory":
                content = (
                    content[:index].decode("utf8")
                    + "memory"
                    + content[index + len("calldata") :].decode("utf8")
                )
            else:
                assert patch_type == "line_removal"
                content = content[:index].decode("utf8") + " // " + content[index:].decode("utf8")

        self._source_codes[contract] = content.decode("utf8")

    def _pragmas(self) -> str:
        """
        Return the required pragmas
        :return:
        """
        ret = ""
        if self._pragma_solidity:
            ret += f"pragma solidity {self._pragma_solidity};\n"
        else:
            # TODO support multiple compiler version
            ret += f"pragma solidity {list(self._compilation_unit.crytic_compile.compilation_units.values())[0].compiler_version.version};\n"

        if self._use_abi_encoder_v2:
            ret += "pragma experimental ABIEncoderV2;\n"
        return ret

    def _export_from_type(
        self,
        t: Type,
        contract: Contract,
        exported: set[str],
        list_contract: set[Contract],
        list_top_level: set[TopLevel],
    ):
        if isinstance(t, UserDefinedType):
            t_type = t.type
            if isinstance(t_type, TopLevel):
                list_top_level.add(t_type)
            elif isinstance(t_type, (EnumContract, StructureContract)):
                if t_type.contract != contract and t_type.contract not in exported:
                    self._export_list_used_contracts(
                        t_type.contract, exported, list_contract, list_top_level
                    )
            else:
                assert isinstance(t.type, Contract)
                if t.type != contract and t.type not in exported:
                    self._export_list_used_contracts(
                        t.type, exported, list_contract, list_top_level
                    )
        elif isinstance(t, MappingType):
            self._export_from_type(t.type_from, contract, exported, list_contract, list_top_level)
            self._export_from_type(t.type_to, contract, exported, list_contract, list_top_level)
        elif isinstance(t, ArrayType):
            self._export_from_type(t.type, contract, exported, list_contract, list_top_level)

    def _export_list_used_contracts(
        self,
        contract: Contract,
        exported: set[str],
        list_contract: set[Contract],
        list_top_level: set[TopLevel],
    ):
        # TODO: investigate why this happen
        if not isinstance(contract, Contract):
            return
        if contract.name in exported:
            return
        exported.add(contract.name)
        for inherited in contract.inheritance:
            self._export_list_used_contracts(inherited, exported, list_contract, list_top_level)

        # Find all the external contracts called
        # High level calls already includes library calls
        # We also filter call to itself to avoid infilite loop
        externals = list({e[0] for e in contract.all_high_level_calls if e[0] != contract})

        for inherited in externals:
            self._export_list_used_contracts(inherited, exported, list_contract, list_top_level)

        for list_libs in contract.using_for.values():
            for lib_candidate_type in list_libs:
                if isinstance(lib_candidate_type, UserDefinedType):
                    lib_candidate = lib_candidate_type.type
                    if isinstance(lib_candidate, Contract):
                        self._export_list_used_contracts(
                            lib_candidate, exported, list_contract, list_top_level
                        )

        # Find all the external contracts use as a base type
        local_vars = []
        for f in contract.functions_declared:
            local_vars += f.variables

        for v in contract.variables + local_vars:
            self._export_from_type(v.type, contract, exported, list_contract, list_top_level)

        for s in contract.structures:
            for elem in s.elems.values():
                self._export_from_type(elem.type, contract, exported, list_contract, list_top_level)

        # Find all convert and "new" operation that can lead to use an external contract
        for f in contract.functions_declared:
            for ir in f.slithir_operations:
                if isinstance(ir, NewContract):
                    if ir.contract_created != contract and ir.contract_created not in exported:
                        self._export_list_used_contracts(
                            ir.contract_created, exported, list_contract, list_top_level
                        )
                if isinstance(ir, TypeConversion):
                    self._export_from_type(
                        ir.type, contract, exported, list_contract, list_top_level
                    )

                for read in ir.read:
                    if isinstance(read, TopLevel):
                        list_top_level.add(read)
                if isinstance(ir, InternalCall) and isinstance(ir.function, FunctionTopLevel):
                    list_top_level.add(ir.function)
                if (
                    isinstance(ir, SolidityCall)
                    and isinstance(ir.function, SolidityCustomRevert)
                    and isinstance(ir.function.custom_error, TopLevel)
                ):
                    list_top_level.add(ir.function.custom_error)

        list_contract.add(contract)

    def _export_contract_with_inheritance(self, contract) -> Export:
        list_contracts: set[Contract] = set()  # will contain contract itself
        list_top_level: set[TopLevel] = set()
        self._export_list_used_contracts(contract, set(), list_contracts, list_top_level)
        path = Path(self._export_path, f"{contract.name}_{uuid.uuid4()}.sol")

        content = ""
        content += self._pragmas()

        for listed_top_level in list_top_level:
            content += self._source_codes_top_level[listed_top_level]
            content += "\n"

        for listed_contract in list_contracts:
            content += self._source_codes[listed_contract]
            content += "\n"

        return Export(filename=path, content=content)

    def _export_most_derived(self) -> list[Export]:
        ret: list[Export] = []
        for contract in self._compilation_unit.contracts_derived:
            ret.append(self._export_contract_with_inheritance(contract))
        return ret

    def _export_all(self) -> list[Export]:
        path = Path(self._export_path, "export.sol")

        content = ""
        content += self._pragmas()

        for top_level_content in self._source_codes_top_level.values():
            content += "\n"
            content += top_level_content
            content += "\n"

        contract_seen = set()
        contract_to_explore = list(self._compilation_unit.contracts)

        # We only need the inheritance order here, as solc can compile
        # a contract that use another contract type (ex: state variable) that he has not seen yet
        while contract_to_explore:
            next_to_explore = contract_to_explore.pop(0)

            if not next_to_explore.inheritance or all(
                father in contract_seen for father in next_to_explore.inheritance
            ):
                content += "\n"
                content += self._source_codes[next_to_explore]
                content += "\n"
                contract_seen.add(next_to_explore)
            else:
                contract_to_explore.append(next_to_explore)

        return [Export(filename=path, content=content)]

    def _export_with_import(self) -> list[Export]:
        exports: list[Export] = []
        for contract in self._compilation_unit.contracts:
            list_contracts: set[Contract] = set()  # will contain contract itself
            list_top_level: set[TopLevel] = set()
            self._export_list_used_contracts(contract, set(), list_contracts, list_top_level)

            if list_top_level:
                logger.info(
                    "Top level objects are not yet supported with the local import flattening"
                )
                for elem in list_top_level:
                    logger.info(f"Missing {elem} for {contract.name}")

            path = Path(self._export_path, f"{contract.name}.sol")

            content = ""
            content += self._pragmas()
            for used_contract in list_contracts:
                if used_contract != contract:
                    content += f"import './{used_contract.name}.sol';\n"
            content += "\n"
            content += self._source_codes[contract]
            content += "\n"
            exports.append(Export(filename=path, content=content))
        return exports

    def export(
        self,
        strategy: Strategy,
        target: str | None = None,
        json: str | None = None,
        zip: str | None = None,
        zip_type: str | None = None,
    ):
        if not self._export_path.exists():
            self._export_path.mkdir(parents=True)

        exports: list[Export] = []
        if target is None:
            if strategy == Strategy.MostDerived:
                exports = self._export_most_derived()
            elif strategy == Strategy.OneFile:
                exports = self._export_all()
            elif strategy == Strategy.LocalImport:
                exports = self._export_with_import()
        else:
            contracts = self._compilation_unit.get_contract_from_name(target)
            if len(contracts) == 0:
                logger.error(f"{target} not found")
                return
            exports = []
            for contract in contracts:
                exports.append(self._export_contract_with_inheritance(contract))

        if json:
            export_as_json(exports, json)

        elif zip:
            save_to_zip(exports, zip, zip_type)

        else:
            save_to_disk(exports)



"""
Module printing summary of the contract
"""

from slither.printers.abstract_printer import AbstractPrinter
from slither.utils.myprettytable import MyPrettyTable


class FunctionSummary(AbstractPrinter):
    ARGUMENT = "function-summary"
    HELP = "Print a summary of the functions"

    WIKI = "https://github.com/trailofbits/slither/wiki/Printer-documentation#function-summary"

    @staticmethod
    def _convert(l):
        if l:
            n = 2
            l = [l[i : i + n] for i in range(0, len(l), n)]
            l = [str(x) for x in l]
            return "\n".join(l)
        return str(l)

    def output(self, _filename):
        """
        _filename is not used
        Args:
            _filename(string)
        """

        all_tables = []
        all_txt = ""

        for c in self.contracts:
            (name, inheritance, var, func_summaries, modif_summaries) = c.get_summary()
            txt = f"\nContract {name}"
            txt += "\nContract vars: " + str(var)
            txt += "\nInheritance:: " + str(inheritance)
            table = MyPrettyTable(
                [
                    "Function",
                    "Visibility",
                    "Modifiers",
                    "Read",
                    "Write",
                    "Internal Calls",
                    "External Calls",
                    "Cyclomatic Complexity",
                ]
            )
            for (
                _c_name,
                f_name,
                visi,
                modifiers,
                read,
                write,
                internal_calls,
                external_calls,
                cyclomatic_complexity,
            ) in func_summaries:
                read = self._convert(sorted(read))
                write = self._convert(sorted(write))
                internal_calls = self._convert(sorted(internal_calls))
                external_calls = self._convert(sorted(external_calls))
                table.add_row(
                    [
                        f_name,
                        visi,
                        sorted(modifiers),
                        read,
                        write,
                        internal_calls,
                        external_calls,
                        cyclomatic_complexity,
                    ]
                )
            txt += "\n \n" + str(table)
            table = MyPrettyTable(
                [
                    "Modifiers",
                    "Visibility",
                    "Read",
                    "Write",
                    "Internal Calls",
                    "External Calls",
                    "Cyclomatic Complexity",
                ]
            )
            for (
                _c_name,
                f_name,
                visi,
                _,
                read,
                write,
                internal_calls,
                external_calls,
                cyclomatic_complexity,
            ) in modif_summaries:
                read = self._convert(sorted(read))
                write = self._convert(sorted(write))
                internal_calls = self._convert(sorted(internal_calls))
                external_calls = self._convert(sorted(external_calls))
                table.add_row(
                    [
                        f_name,
                        visi,
                        read,
                        write,
                        internal_calls,
                        external_calls,
                        cyclomatic_complexity,
                    ]
                )
            txt += "\n\n" + str(table)
            txt += "\n"
            self.info(txt)

            all_tables.append((name, table))
            all_txt += txt

        res = self.generate_output(all_txt)
        for name, table in all_tables:
            res.add_pretty_table(table, name)

        return res



from pathlib import Path
from slither import Slither

# % solc functions_ids.sol --hashes
# ======= functions_ids.sol:C =======
# Function signatures:
# 0dbe671f: a()
# 4a1f689d: a_array(uint256)
# 98fc2aa5: arrayOfMappings(uint256,uint256)
# 4ea7a557: b_mapping_of_array(address,uint256)
# 3c0af344: contractMap(address)
# 20969954: function_with_array(((uint256,uint256),uint256)[],(uint256,(uint256,uint256)))
# 1c039831: function_with_struct(((uint256,uint256),uint256))
# 37e66bae: mapping_of_double_array_of_struct(address,uint256,uint256)
# f29872a8: multiDimensionalArray(uint256,uint256)
# 9539e3c8: normalMappingArrayField(uint256,uint256)
# 87c3dbb6: outer()
# df201a46: simple()
# 5a20851f: stateMap(uint16)

# {"contracts":{"functions_ids.sol:C":{"hashes":{"a()":"0dbe671f","a_array(uint256)":"4a1f689d","arrayOfMappings(uint256,uint256)":"98fc2aa5","b_mapping_of_array(address,uint256)":"4ea7a557","contractMap(address)":"3c0af344","function_with_array(((uint256,uint256),uint256)[],(uint256,(uint256,uint256)))":"20969954","function_with_struct(((uint256,uint256),uint256))":"1c039831","mapping_of_double_array_of_struct(address,uint256,uint256)":"37e66bae","multiDimensionalArray(uint256,uint256)":"f29872a8","normalMappingArrayField(uint256,uint256)":"9539e3c8","outer()":"87c3dbb6","simple()":"df201a46","stateMap(uint16)":"5a20851f"}},"functions_ids.sol:Contract":{"hashes":{}}},"version":"0.7.0+commit.9e61f92b.Darwin.appleclang"}
from slither.utils.function import get_function_id

signatures = {
    "a()": "0dbe671f",
    "a_array(uint256)": "4a1f689d",
    "arrayOfMappings(uint256,uint256)": "98fc2aa5",
    "b_mapping_of_array(address,uint256)": "4ea7a557",
    "contractMap(address)": "3c0af344",
    "function_with_array(((uint256,uint256),uint256)[],(uint256,(uint256,uint256)))": "20969954",
    "function_with_struct(((uint256,uint256),uint256))": "1c039831",
    "mapping_of_double_array_of_struct(address,uint256,uint256)": "37e66bae",
    "multiDimensionalArray(uint256,uint256)": "f29872a8",
    "normalMappingArrayField(uint256,uint256)": "9539e3c8",
    "outer()": "87c3dbb6",
    "simple()": "df201a46",
    "stateMap(uint16)": "5a20851f",
}

TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"


def test_functions_ids(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.7.0")
    file = Path(TEST_DATA_DIR, "functions_ids.sol").as_posix()
    sl = Slither(file, solc=solc_path)
    contracts_c = sl.get_contract_from_name("C")
    assert len(contracts_c) == 1
    contract_c = contracts_c[0]

    for sig, hashes in signatures.items():
        func = contract_c.get_function_from_signature(sig)
        if not func:
            var_name = sig[: sig.find("(")]
            var = contract_c.get_state_variable_from_name(var_name)
            assert var
            assert get_function_id(var.solidity_signature) == int(hashes, 16)
        else:
            assert get_function_id(func.solidity_signature) == int(hashes, 16)



from slither.core.cfg.node import NodeType, Node
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.slithir.operations import InternalCall
from slither.core.declarations import SolidityVariableComposed, Contract
from slither.utils.output import Output
from slither.slithir.variables.constant import Constant
from slither.core.variables import Variable
from slither.core.expressions.literal import Literal

Result = list[tuple[Node, list[str]]]


def detect_msg_value_in_loop(contract: Contract) -> Result:
    results: Result = []
    for f in contract.functions_entry_points:
        if f.is_implemented and f.payable:
            msg_value_in_loop(f.entry_point, 0, [], [], results)
    return results


def msg_value_in_loop(
    node: Node | None,
    in_loop_counter: int,
    visited: list[Node],
    calls_stack: list[str],
    results: Result,
) -> None:
    if node is None:
        return

    if node in visited:
        return
    # shared visited
    visited.append(node)

    if node.type == NodeType.STARTLOOP:
        in_loop_counter += 1
    elif node.type == NodeType.ENDLOOP:
        in_loop_counter -= 1

    for ir in node.irs:
        if in_loop_counter > 0 and SolidityVariableComposed("msg.value") in ir.read:
            # If we find a conditional expression with msg.value and is compared to 0 we don't report it
            if ir.node.is_conditional() and SolidityVariableComposed("msg.value") in ir.read:
                compared_to = (
                    ir.read[1]
                    if ir.read[0] == SolidityVariableComposed("msg.value")
                    else ir.read[0]
                )
                if (isinstance(compared_to, Constant) and compared_to.value == 0) or (
                    isinstance(compared_to, Variable)
                    and isinstance(compared_to.expression, Literal)
                    and str(compared_to.expression.value) == "0"
                ):
                    continue
            results.append((ir.node, calls_stack.copy()))
        if isinstance(ir, (InternalCall)):
            calls_stack.append(node.function.canonical_name)
            msg_value_in_loop(
                ir.function.entry_point, in_loop_counter, visited, calls_stack, results
            )
            calls_stack.pop()

    for son in node.sons:
        msg_value_in_loop(son, in_loop_counter, visited, calls_stack, results)


class MsgValueInLoop(AbstractDetector):
    """
    Detect the use of msg.value inside a loop
    """

    ARGUMENT = "msg-value-loop"
    HELP = "msg.value inside a loop"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation/#msgvalue-inside-a-loop"

    WIKI_TITLE = "`msg.value` inside a loop"
    WIKI_DESCRIPTION = "Detect the use of `msg.value` inside a loop."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract MsgValueInLoop{

    mapping (address => uint256) balances;

    function bad(address[] memory receivers) public payable {
        for (uint256 i=0; i < receivers.length; i++) {
            balances[receivers[i]] += msg.value;
        }
    }

}
```
"""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = """
Provide an explicit array of amounts alongside the receivers array, and check that the sum of all amounts matches `msg.value`.
"""

    def _detect(self) -> list[Output]:
        """"""
        results: list[Output] = []
        for c in self.compilation_unit.contracts_derived:
            values = detect_msg_value_in_loop(c)
            for node, calls_stack in values:
                func = node.function

                info: DETECTOR_INFO = [func, " use msg.value in a loop: ", node, "\n"]

                if len(calls_stack) > 0:
                    info.append("\tCalls stack containing the loop:\n")
                    for call in calls_stack:
                        info.extend(["\t\t", call, "\n"])

                res = self.generate_result(info)
                results.append(res)

        return results



from slither.tools.properties.properties.properties import (
    PropertyType,
    PropertyReturn,
    Property,
    PropertyCaller,
)

ERC20_NotMintable = [
    Property(
        name="crytic_supply_constant_ERC20PropertiesNotMintable()",
        description="The total supply does not increase.",
        content="""
\t\treturn initialTotalSupply >= this.totalSupply();""",
        type=PropertyType.MEDIUM_SEVERITY,
        return_type=PropertyReturn.SUCCESS,
        is_unit_test=True,
        is_property_test=True,
        caller=PropertyCaller.ANY,
    ),
]



"""
Tool to read on-chain storage from EVM
"""

import json
import argparse

from crytic_compile import cryticparser

from slither import Slither
from slither.exceptions import SlitherError
from slither.tools.read_storage.read_storage import SlitherReadStorage, RpcInfo


def parse_args() -> argparse.Namespace:
    """Parse the underlying arguments for the program.
    Returns:
        The arguments for the program.
    """
    parser = argparse.ArgumentParser(
        description="Read a variable's value from storage for a deployed contract",
        usage=(
            "\nTo retrieve a single variable's value:\n"
            + "\tslither-read-storage $TARGET address --variable-name $NAME\n"
            + "To retrieve a contract's storage layout:\n"
            + "\tslither-read-storage $TARGET address --contract-name $NAME --json storage_layout.json\n"
            + "To retrieve a contract's storage layout and values:\n"
            + "\tslither-read-storage $TARGET address --contract-name $NAME --json storage_layout.json --value\n"
            + "TARGET can be a contract address or project directory"
        ),
    )

    parser.add_argument(
        "contract_source",
        help="The deployed contract address if verified on etherscan. Prepend project directory for unverified contracts.",
        nargs="+",
    )

    parser.add_argument(
        "--variable-name",
        help="The name of the variable whose value will be returned.",
        default=None,
    )

    parser.add_argument("--rpc-url", help="An endpoint for web3 requests.")

    parser.add_argument(
        "--key",
        help="The key/ index whose value will be returned from a mapping or array.",
        default=None,
    )

    parser.add_argument(
        "--deep-key",
        help="The key/ index whose value will be returned from a deep mapping or multidimensional array.",
        default=None,
    )

    parser.add_argument(
        "--struct-var",
        help="The name of the variable whose value will be returned from a struct.",
        default=None,
    )

    parser.add_argument(
        "--storage-address",
        help="The address of the storage contract (if a proxy pattern is used).",
        default=None,
    )

    parser.add_argument(
        "--contract-name",
        help="The name of the logic contract.",
        default=None,
    )

    parser.add_argument(
        "--json",
        action="store",
        help="Save the result in a JSON file.",
    )

    parser.add_argument(
        "--value",
        action="store_true",
        help="Toggle used to include values in output.",
    )

    parser.add_argument(
        "--table",
        action="store_true",
        help="Print table view of storage layout",
    )

    parser.add_argument(
        "--silent",
        action="store_true",
        help="Silence log outputs",
    )

    parser.add_argument("--max-depth", help="Max depth to search in data structure.", default=20)

    parser.add_argument(
        "--block",
        help="The block number to read storage from. Requires an archive node to be provided as the RPC url.",
        default="latest",
    )

    parser.add_argument(
        "--unstructured",
        action="store_true",
        help="Include unstructured storage slots",
    )

    cryticparser.init(parser)

    return parser.parse_args()


def main() -> None:
    args = parse_args()

    if len(args.contract_source) == 2:
        # Source code is file.sol or project directory
        source_code, target = args.contract_source
        slither = Slither(source_code, **vars(args))
    else:
        # Source code is published and retrieved via etherscan
        target = args.contract_source[0]
        slither = Slither(target, **vars(args))

    if args.contract_name:
        contracts = slither.get_contract_from_name(args.contract_name)
        if len(contracts) == 0:
            raise SlitherError(f"Contract {args.contract_name} not found.")
    else:
        contracts = slither.contracts

    rpc_info = None
    if args.rpc_url:
        valid = ["latest", "earliest", "pending", "safe", "finalized"]
        block = args.block if args.block in valid else int(args.block)
        rpc_info = RpcInfo(args.rpc_url, block)

    srs = SlitherReadStorage(contracts, args.max_depth, rpc_info)
    srs.unstructured = bool(args.unstructured)
    # Remove target prefix e.g. rinkeby:0x0 -> 0x0.
    address = target[target.find(":") + 1 :]
    # Default to implementation address unless a storage address is given.
    if not args.storage_address:
        args.storage_address = address
    srs.storage_address = args.storage_address

    if args.variable_name:
        # Use a lambda func to only return variables that have same name as target.
        # x is a tuple (`Contract`, `StateVariable`).
        srs.get_all_storage_variables(lambda x: bool(x[1].name == args.variable_name))
        srs.get_target_variables(**vars(args))
    else:
        srs.get_all_storage_variables()
        srs.get_storage_layout()

    # To retrieve slot values an rpc url is required.
    if args.value:
        assert args.rpc_url
        srs.walk_slot_info(srs.get_slot_values)

    if args.table:
        srs.walk_slot_info(srs.convert_slot_info_to_rows)
        print(srs.table)

    if args.json:
        with open(args.json, "w", encoding="utf8") as file:
            slot_infos_json = srs.to_json()
            json.dump(slot_infos_json, file, indent=4)


if __name__ == "__main__":
    main()



import sys
from slither.slither import Slither

if len(sys.argv) != 2:
    print("python functions_called.py functions_called.sol")
    sys.exit(-1)

# Init slither
slither = Slither(sys.argv[1])

# Get the contract
contracts = slither.get_contract_from_name("Contract")
assert len(contracts) == 1
contract = contracts[0]

# Get the variable
entry_point = contract.get_function_from_signature("entry_point()")
assert entry_point

all_calls = entry_point.all_internal_calls()

all_calls_formated = [f.canonical_name for f in all_calls]

# Print the result
print(f"From entry_point the functions reached are {all_calls_formated}")



import logging
from pathlib import Path

from crytic_compile import crytic_compile

from slither.tools.doctor.utils import snip_section
from slither.utils.colors import red, yellow, green


def detect_platform(project: str, **kwargs) -> None:
    path = Path(project)
    if path.is_file():
        print(
            yellow(
                f"{project!r} is a file. Using it as target will manually compile your code with solc and _not_ use a compilation framework. Is that what you meant to do?"
            )
        )
        return

    print(f"Trying to detect project type for {project!r}")

    supported_platforms = crytic_compile.get_platforms()
    skip_platforms = {"solc", "solc-json", "archive", "standard", "etherscan"}
    detected_platforms = {
        platform.NAME: platform.is_supported(project, **kwargs)
        for platform in supported_platforms
        if platform.NAME.lower() not in skip_platforms
    }
    platform_qty = len([platform for platform, state in detected_platforms.items() if state])

    print("Is this project using...")
    for platform, state in detected_platforms.items():
        print(f"    =>  {platform + '?':<15}{(state and green('Yes')) or red('No')}")
    print()

    if platform_qty == 0:
        print(red("No platform was detected! This doesn't sound right."))
        print(
            yellow(
                "Are you trying to analyze a folder with standalone solidity files, without using a compilation framework? If that's the case, then this is okay."
            )
        )
    elif platform_qty > 1:
        print(red("More than one platform was detected! This doesn't sound right."))
        print(
            red("Please use `--compile-force-framework` in Slither to force the correct framework.")
        )
    else:
        print(green("A single platform was detected."), yellow("Is it the one you expected?"))


def compile_project(project: str, **kwargs):
    print("Invoking crytic-compile on the project, please wait...")

    try:
        crytic_compile.CryticCompile(project, **kwargs)
    except Exception as e:
        with snip_section("Project compilation failed :( The following error was generated:"):
            logging.exception(e)



"""
Module detecting storage signed integer array bug
"""

from slither.core.declarations import Function, Contract
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    make_solc_versions,
    DETECTOR_INFO,
)
from slither.core.cfg.node import NodeType, Node
from slither.core.solidity_types import ArrayType
from slither.core.solidity_types.elementary_type import Int, ElementaryType
from slither.core.variables.local_variable import LocalVariable
from slither.core.variables.state_variable import StateVariable
from slither.slithir.operations import Operation, OperationWithLValue
from slither.slithir.operations.assignment import Assignment
from slither.slithir.operations.init_array import InitArray
from slither.utils.output import Output


class StorageSignedIntegerArray(AbstractDetector):
    """
    Storage signed integer array
    """

    ARGUMENT = "storage-array"
    HELP = "Signed storage integer array compiler bug"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = (
        "https://github.com/crytic/slither/wiki/Detector-Documentation#storage-signed-integer-array"
    )
    WIKI_TITLE = "Storage Signed Integer Array"

    # region wiki_description
    WIKI_DESCRIPTION = """`solc` versions `0.4.7`-`0.5.9` contain [a compiler bug](https://blog.ethereum.org/2019/06/25/solidity-storage-array-bugs)
leading to incorrect values in signed integer arrays."""
    # endregion wiki_description

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract A {
	int[3] ether_balances; // storage signed integer array
	function bad0() private {
		// ...
		ether_balances = [-1, -1, -1];
		// ...
	}
}
```
`bad0()` uses a (storage-allocated) signed integer array state variable to store the ether balances of three accounts.
`-1` is supposed to indicate uninitialized values but the Solidity bug makes these as `1`, which could be exploited by the accounts.
"""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Use a compiler version >= `0.5.10`."

    VULNERABLE_SOLC_VERSIONS = make_solc_versions(4, 7, 25) + make_solc_versions(5, 0, 9)

    @staticmethod
    def _is_vulnerable_type(ir: Operation) -> bool:
        """
        Detect if the IR lvalue is a vulnerable type
        Must be a storage allocation, and an array of Int
        Assume the IR is a InitArray, or an Assignement to an ArrayType
        """
        # Storage allocation
        # Base type is signed integer
        if not isinstance(ir, OperationWithLValue):
            return False

        return (
            (
                isinstance(ir.lvalue, StateVariable)
                or (isinstance(ir.lvalue, LocalVariable) and ir.lvalue.is_storage)
            )
            and isinstance(ir.lvalue.type.type, ElementaryType)  # type: ignore
            and ir.lvalue.type.type.type in Int  # type: ignore
        )

    def detect_storage_signed_integer_arrays(
        self, contract: Contract
    ) -> set[tuple[Function, Node]]:
        """
        Detects and returns all nodes with storage-allocated signed integer array init/assignment
        :param contract: Contract to detect within
        :return: A list of tuples with (function, node) where function node has storage-allocated signed integer array init/assignment
        """
        # Create our result set.
        results: set[tuple[Function, Node]] = set()

        # Loop for each function and modifier.
        for function in contract.functions_and_modifiers_declared:
            # Loop for every node in this function, looking for storage-allocated
            # signed integer array initializations/assignments
            for node in function.nodes:
                if node.type == NodeType.EXPRESSION:
                    for ir in node.irs:
                        # Storage-allocated signed integer array initialization expression
                        if isinstance(ir, InitArray) and self._is_vulnerable_type(ir):
                            results.add((function, node))
                        # Assignment expression with lvalue being a storage-allocated signed integer array and
                        # rvalue being a signed integer array of different base type than lvalue
                        if (
                            isinstance(ir, Assignment)
                            and isinstance(ir.lvalue.type, ArrayType)
                            and self._is_vulnerable_type(ir)
                            # Base type is signed integer and lvalue base type is different from rvalue base type
                            and ir.lvalue.type.type != ir.rvalue.type.type
                        ):
                            results.add((function, node))

        # Return the resulting set of tuples
        return results

    def _detect(self) -> list[Output]:
        """
        Detect storage signed integer array init/assignment
        """
        results = []
        for contract in self.contracts:
            storage_signed_integer_arrays = self.detect_storage_signed_integer_arrays(contract)
            for function, node in storage_signed_integer_arrays:
                contract_info: DETECTOR_INFO = ["Contract ", contract, " \n"]
                function_info: DETECTOR_INFO = ["\t- Function ", function, "\n"]
                node_info: DETECTOR_INFO = [
                    "\t\t- ",
                    node,
                    " has a storage signed integer array assignment\n",
                ]
                res = self.generate_result(contract_info + function_info + node_info)
                results.append(res)

        return results



import logging
from slither.exceptions import SlitherError

logger = logging.getLogger("ConvertToEVM")


def load_evm_cfg_builder():
    try:
        # Avoiding the addition of evm_cfg_builder as permanent dependency

        from evm_cfg_builder.cfg import CFG

        return CFG
    except ImportError:
        logger.error("To use evm features, you need to install evm-cfg-builder")
        logger.error("Documentation: https://github.com/crytic/evm_cfg_builder")
        logger.error("Installation: pip install evm-cfg-builder")

        raise SlitherError("evm-cfg-builder not installed.")



from slither.core.declarations import Function
from slither.slithir.operations.call import Call
from slither.slithir.operations.lvalue import OperationWithLValue
from slither.core.variables.variable import Variable
from slither.core.declarations.solidity_variables import SolidityVariable

from slither.slithir.variables.constant import Constant
from slither.core.variables.local_variable import LocalVariable
from slither.slithir.variables.local_variable import LocalIRVariable
from slither.slithir.variables.temporary import TemporaryVariable
from slither.slithir.variables.temporary_ssa import TemporaryVariableSSA
from slither.slithir.variables.tuple import TupleVariable
from slither.slithir.variables.tuple_ssa import TupleVariableSSA


class LowLevelCall(Call, OperationWithLValue):
    """
    High level message call
    """

    def __init__(
        self,
        destination: LocalVariable | LocalIRVariable | TemporaryVariableSSA | TemporaryVariable,
        function_name: Constant,
        nbr_arguments: int,
        result: TupleVariable | TupleVariableSSA,
        type_call: str,
    ) -> None:
        assert isinstance(destination, (Variable, SolidityVariable))
        assert isinstance(function_name, Constant)
        super().__init__()
        self._destination = destination
        self._function_name = function_name
        self._nbr_arguments = nbr_arguments
        self._type_call = type_call
        self._lvalue = result
        self._callid = None  # only used if gas/value != 0

        self._call_value = None
        self._call_gas = None

    @property
    def call_id(self):
        return self._callid

    @call_id.setter
    def call_id(self, c):
        self._callid = c

    @property
    def call_value(self):
        return self._call_value

    @call_value.setter
    def call_value(self, v):
        self._call_value = v

    @property
    def call_gas(self):
        return self._call_gas

    @call_gas.setter
    def call_gas(self, v):
        self._call_gas = v

    @property
    def read(
        self,
    ) -> list[
        LocalIRVariable | Constant | LocalVariable | TemporaryVariableSSA | TemporaryVariable
    ]:
        all_read = [self.destination, self.call_gas, self.call_value] + self.arguments
        # remove None
        return self._unroll([x for x in all_read if x])

    def can_reenter(self, _callstack: list[Function | Variable] | None = None) -> bool:
        """
        Must be called after slithIR analysis pass
        :return: bool
        """
        if self.function_name == "staticcall":
            return False
        return True

    def can_send_eth(self) -> bool:
        """
        Must be called after slithIR analysis pass
        :return: bool
        """
        return self._call_value is not None

    @property
    def destination(
        self,
    ) -> LocalVariable | LocalIRVariable | TemporaryVariableSSA | TemporaryVariable:
        return self._destination

    @property
    def function_name(self) -> Constant:
        return self._function_name

    @property
    def nbr_arguments(self) -> int:
        return self._nbr_arguments

    @property
    def type_call(self) -> str:
        return self._type_call

    def __str__(self):
        value = ""
        gas = ""
        if self.call_value:
            value = f"value:{self.call_value}"
        if self.call_gas:
            gas = f"gas:{self.call_gas}"
        arguments = []
        if self.arguments:
            arguments = self.arguments
        return_type = self.lvalue.type

        if return_type and isinstance(return_type, list):
            return_type = ",".join(str(x) for x in return_type)

        txt = "{}({}) = LOW_LEVEL_CALL, dest:{}, function:{}, arguments:{} {} {}"
        return txt.format(
            self.lvalue,
            return_type,
            self.destination,
            self.function_name,
            [str(x) for x in arguments],
            value,
            gas,
        )



from slither.core.variables.structure_variable import StructureVariable
from slither.vyper_parsing.type_parsing import parse_type
from slither.vyper_parsing.ast.types import AnnAssign


class StructureVariableVyper:
    def __init__(self, variable: StructureVariable, variable_data: AnnAssign):
        self._variable: StructureVariable = variable
        self._variable.name = variable_data.target.id
        self._elem_to_parse = variable_data.annotation

    @property
    def underlying_variable(self) -> StructureVariable:
        return self._variable

    def analyze(self, contract) -> None:
        self._variable.type = parse_type(self._elem_to_parse, contract)



from slither.tools.upgradeability.checks.abstract_checks import classification_txt, AbstractCheck
from slither.utils.myprettytable import MyPrettyTable


def output_wiki(detector_classes: list[type[AbstractCheck]], filter_wiki: str) -> None:
    # Sort by impact, confidence, and name
    detectors_list = sorted(
        detector_classes, key=lambda element: (element.IMPACT, element.ARGUMENT)
    )

    for detector in detectors_list:
        if filter_wiki not in detector.WIKI:
            continue
        argument = detector.ARGUMENT
        impact = classification_txt[detector.IMPACT]
        title = detector.WIKI_TITLE
        description = detector.WIKI_DESCRIPTION
        exploit_scenario = detector.WIKI_EXPLOIT_SCENARIO
        recommendation = detector.WIKI_RECOMMENDATION

        print(f"\n## {title}")
        print("### Configuration")
        print(f"* Check: `{argument}`")
        print(f"* Severity: `{impact}`")
        print("\n### Description")
        print(description)
        if exploit_scenario:
            print("\n### Exploit Scenario:")
            print(exploit_scenario)
        print("\n### Recommendation")
        print(recommendation)


def output_detectors(detector_classes: list[type[AbstractCheck]]) -> None:
    detectors_list = []
    for detector in detector_classes:
        argument = detector.ARGUMENT
        help_info = detector.HELP
        impact = detector.IMPACT
        require_proxy = detector.REQUIRE_PROXY
        require_v2 = detector.REQUIRE_CONTRACT_V2
        detectors_list.append((argument, help_info, impact, require_proxy, require_v2))
    table = MyPrettyTable(["Num", "Check", "What it Detects", "Impact", "Proxy", "Contract V2"])

    # Sort by impact, confidence, and name
    detectors_list = sorted(detectors_list, key=lambda element: (element[2], element[0]))
    idx = 1
    for argument, help_info, impact, proxy, v2 in detectors_list:
        table.add_row(
            [
                str(idx),
                argument,
                help_info,
                classification_txt[impact],
                "X" if proxy else "",
                "X" if v2 else "",
            ]
        )
        idx = idx + 1
    print(table)


def output_to_markdown(detector_classes: list[type[AbstractCheck]], _filter_wiki: str) -> None:
    def extract_help(cls: type[AbstractCheck]) -> str:
        if cls.WIKI == "":
            return cls.HELP
        return f"[{cls.HELP}]({cls.WIKI})"

    detectors_list = []
    for detector in detector_classes:
        argument = detector.ARGUMENT
        help_info = extract_help(detector)
        impact = detector.IMPACT
        require_proxy = detector.REQUIRE_PROXY
        require_v2 = detector.REQUIRE_CONTRACT_V2
        detectors_list.append((argument, help_info, impact, require_proxy, require_v2))

    # Sort by impact, confidence, and name
    detectors_list = sorted(detectors_list, key=lambda element: (element[2], element[0]))
    idx = 1
    for argument, help_info, impact, proxy, v2 in detectors_list:
        print(
            f"{idx} | `{argument}` | {help_info} | {classification_txt[impact]} | {'X' if proxy else ''} | {'X' if v2 else ''}"
        )
        idx = idx + 1


def output_detectors_json(
    detector_classes: list[type[AbstractCheck]],
) -> list[dict[str, str | int]]:
    detectors_list = []
    for detector in detector_classes:
        argument = detector.ARGUMENT
        help_info = detector.HELP
        impact = detector.IMPACT
        wiki_url = detector.WIKI
        wiki_description = detector.WIKI_DESCRIPTION
        wiki_exploit_scenario = detector.WIKI_EXPLOIT_SCENARIO
        wiki_recommendation = detector.WIKI_RECOMMENDATION
        detectors_list.append(
            (
                argument,
                help_info,
                impact,
                wiki_url,
                wiki_description,
                wiki_exploit_scenario,
                wiki_recommendation,
            )
        )

    # Sort by impact, confidence, and name
    detectors_list = sorted(detectors_list, key=lambda element: (element[2], element[0]))
    idx = 1
    table: list[dict[str, str | int]] = []
    for (
        argument,
        help_info,
        impact,
        wiki_url,
        description,
        exploit,
        recommendation,
    ) in detectors_list:
        table.append(
            {
                "index": idx,
                "check": argument,
                "title": help_info,
                "impact": classification_txt[impact],
                "wiki_url": wiki_url,
                "description": description,
                "exploit_scenario": exploit,
                "recommendation": recommendation,
            }
        )
        idx = idx + 1
    return table



import argparse
import logging
import sys

from crytic_compile import cryticparser

from slither.tools.doctor.utils import report_section
from slither.tools.doctor.checks import ALL_CHECKS


def parse_args() -> argparse.Namespace:
    """
    Parse the underlying arguments for the program.
    :return: Returns the arguments for the program.
    """
    parser = argparse.ArgumentParser(
        description="Troubleshoot running Slither on your project",
        usage="slither-doctor project",
    )

    parser.add_argument("project", help="The codebase to be tested.")

    # Add default arguments from crytic-compile
    cryticparser.init(parser)

    return parser.parse_args()


def main() -> None:
    # log on stdout to keep output in order
    logging.basicConfig(stream=sys.stdout, level=logging.DEBUG, force=True)

    args = parse_args()
    kwargs = vars(args)

    for check in ALL_CHECKS:
        with report_section(check.title):
            check.function(**kwargs)


if __name__ == "__main__":
    main()



from typing import TYPE_CHECKING

from slither.core.declarations import Contract, Enum, SolidityVariable, Function
from slither.core.variables.variable import Variable
from slither.core.variables.top_level_variable import TopLevelVariable

if TYPE_CHECKING:
    from slither.core.cfg.node import Node


class ReferenceVariable(Variable):
    def __init__(self, node: "Node", index: int | None = None) -> None:
        super().__init__()
        if index is None:
            self._index = node.compilation_unit.counter_slithir_reference
            node.compilation_unit.counter_slithir_reference += 1
        else:
            self._index = index
        self._points_to = None
        self._node = node

    @property
    def node(self) -> "Node":
        return self._node

    @property
    def index(self):
        return self._index

    @index.setter
    def index(self, idx):
        self._index = idx

    @property
    def points_to(self):
        """
        Return the variable pointer by the reference
        It is the left member of a Index or Member operator
        """
        return self._points_to

    @points_to.setter
    def points_to(self, points_to):
        # Can only be a rvalue of
        # Member or Index operator

        from slither.slithir.utils.utils import is_valid_lvalue

        assert is_valid_lvalue(points_to) or isinstance(
            points_to, (SolidityVariable, Contract, Enum, TopLevelVariable)
        )

        self._points_to = points_to

    @property
    def points_to_origin(self):
        points = self.points_to
        while isinstance(points, ReferenceVariable):
            points = points.points_to
        return points

    @property
    def name(self) -> str:
        return f"REF_{self.index}"

    # overide of core.variables.variables
    # reference can have Function has a type
    # to handle the function selector
    def set_type(self, t) -> None:
        if not isinstance(t, Function):
            super().set_type(t)
        else:
            self._type = t

    def __str__(self) -> str:
        return self.name



from slither.slithir.operations.call import Call
from slither.slithir.operations.lvalue import OperationWithLValue

from slither.slithir.utils.utils import is_valid_lvalue

from slither.core.declarations.structure import Structure
from slither.core.declarations.structure_contract import StructureContract
from slither.slithir.variables.constant import Constant
from slither.slithir.variables.temporary import TemporaryVariable
from slither.slithir.variables.temporary_ssa import TemporaryVariableSSA


class NewStructure(Call, OperationWithLValue):
    def __init__(
        self,
        structure: StructureContract,
        lvalue: TemporaryVariableSSA | TemporaryVariable,
        names: list[str] | None = None,
    ) -> None:
        """
        #### Parameters
        names -
            For calls of the form f({argName1 : arg1, ...}), the names of parameters listed in call order.
            Otherwise, None.
        """
        super().__init__(names=names)
        assert isinstance(structure, Structure)
        assert is_valid_lvalue(lvalue)
        self._structure = structure
        # todo create analyze to add the contract instance
        self._lvalue = lvalue

    @property
    def read(self) -> list[TemporaryVariableSSA | TemporaryVariable | Constant]:
        return self._unroll(self.arguments)

    @property
    def structure(self) -> StructureContract:
        return self._structure

    @property
    def structure_name(self):
        return self.structure.name

    def __str__(self):
        args = [str(a) for a in self.arguments]
        lvalue = self.lvalue
        return f"{lvalue}({lvalue.type}) = new {self.structure_name}({','.join(args)})"



from typing import Any

from slither.core.declarations import Function
from slither.core.variables.variable import Variable
from slither.slithir.operations.operation import Operation
from slither.slithir.utils.utils import is_valid_rvalue, RVALUE
from slither.slithir.variables.tuple import TupleVariable


class Return(Operation):
    """
    Return
    Only present as last operation in RETURN node
    """

    def __init__(self, values: RVALUE | TupleVariable | Function | list[RVALUE] | None) -> None:
        # Note: Can return None
        # ex: return call()
        # where call() dont return
        self._values: list[RVALUE | TupleVariable | Function]
        if not isinstance(values, list):
            assert (
                is_valid_rvalue(values)
                or isinstance(values, (TupleVariable, Function))
                or values is None
            )
            if values is None:
                self._values = []
            else:
                self._values = [values]
        else:
            # Remove None
            # Prior Solidity 0.5
            # return (0,)
            # was valid for returns(uint)
            self._values = [v for v in values if v is not None]
            self._valid_value(self._values)
        super().__init__()

    def _valid_value(self, value: Any) -> bool:
        if isinstance(value, list):
            assert all(self._valid_value(v) for v in value)
        else:
            assert is_valid_rvalue(value) or isinstance(value, (TupleVariable, Function))
        return True

    @property
    def read(self) -> list[Variable]:
        return self._unroll(self.values)

    @property
    def values(self) -> list[Variable]:
        return self._unroll(self._values)

    def __str__(self) -> str:
        return f"RETURN {','.join([f'{x}' for x in self.values])}"



import sys

from slither import Slither
from slither.analyses.data_dependency.data_dependency import (
    is_dependent,
    is_tainted,
)
from slither.core.declarations.solidity_variables import SolidityVariableComposed

if len(sys.argv) != 2:
    print("Usage: python data_dependency.py file.sol")
    sys.exit(-1)

slither = Slither(sys.argv[1])

contracts = slither.get_contract_from_name("Simple")
assert len(contracts) == 1
contract = contracts[0]
destination = contract.get_state_variable_from_name("destination")
source = contract.get_state_variable_from_name("source")
assert source
assert destination

print(f"{source} is dependent of {destination}: {is_dependent(source, destination, contract)}")
assert not is_dependent(source, destination, contract)
print(f"{destination} is dependent of {source}: {is_dependent(destination, source, contract)}")
assert is_dependent(destination, source, contract)
print(f"{source} is tainted {is_tainted(source, contract)}")
assert not is_tainted(source, contract)
print(f"{destination} is tainted {is_tainted(destination, contract)}")
assert is_tainted(destination, contract)

contracts = slither.get_contract_from_name("Reference")
assert len(contracts) == 1
contract = contracts[0]
destination = contract.get_state_variable_from_name("destination")
assert destination
source = contract.get_state_variable_from_name("source")
assert source

print("Reference contract")
print(f"{source} is dependent of {destination}: {is_dependent(source, destination, contract)}")
assert not is_dependent(source, destination, contract)
print(f"{destination} is dependent of {source}: {is_dependent(destination, source, contract)}")
assert is_dependent(destination, source, contract)
print(f"{source} is tainted {is_tainted(source, contract)}")
assert not is_tainted(source, contract)
print(f"{destination} is tainted {is_tainted(destination, contract)}")
assert is_tainted(destination, contract)

destination_indirect_1 = contract.get_state_variable_from_name("destination_indirect_1")
assert destination_indirect_1
print(f"{destination_indirect_1} is tainted {is_tainted(destination_indirect_1, contract)}")
assert is_tainted(destination_indirect_1, contract)
destination_indirect_2 = contract.get_state_variable_from_name("destination_indirect_2")
assert destination_indirect_2
print(f"{destination_indirect_2} is tainted {is_tainted(destination_indirect_2, contract)}")
assert is_tainted(destination_indirect_2, contract)

print("SolidityVar contract")

contracts = slither.get_contract_from_name("SolidityVar")
assert len(contracts) == 1
contract = contracts[0]
addr_1 = contract.get_state_variable_from_name("addr_1")
assert addr_1
addr_2 = contract.get_state_variable_from_name("addr_2")
assert addr_2
msgsender = SolidityVariableComposed("msg.sender")
print(f"{addr_1} is dependent of {msgsender}: {is_dependent(addr_1, msgsender, contract)}")
assert is_dependent(addr_1, msgsender, contract)
print(f"{addr_2} is dependent of {msgsender}: {is_dependent(addr_2, msgsender, contract)}")
assert not is_dependent(addr_2, msgsender, contract)


print("Intermediate contract")
contracts = slither.get_contract_from_name("Intermediate")
assert len(contracts) == 1
contract = contracts[0]
destination = contract.get_state_variable_from_name("destination")
assert destination
source = contract.get_state_variable_from_name("source")
assert source

print(f"{destination} is dependent of {source}: {is_dependent(destination, source, contract)}")
assert is_dependent(destination, source, contract)

print("Base Derived contract")
contracts = slither.get_contract_from_name("Base")
assert len(contracts) == 1
contract = contracts[0]
contract_derived = slither.get_contract_from_name("Derived")[0]
destination = contract.get_state_variable_from_name("destination")
source = contract.get_state_variable_from_name("source")
assert destination
assert source

print(f"{destination} is dependent of {source}: {is_dependent(destination, source, contract)}")
assert not is_dependent(destination, source, contract)
print(
    f"{destination} is dependent of {source}: {is_dependent(destination, source, contract_derived)}"
)
assert is_dependent(destination, source, contract_derived)

print("PropagateThroughArguments contract")
contracts = slither.get_contract_from_name("PropagateThroughArguments")
assert len(contracts) == 1
contract = contracts[0]
var_tainted = contract.get_state_variable_from_name("var_tainted")
assert var_tainted
var_not_tainted = contract.get_state_variable_from_name("var_not_tainted")
assert var_not_tainted
var_dependant = contract.get_state_variable_from_name("var_dependant")
assert var_dependant

f = contract.get_function_from_signature("f(uint256)")
assert f
user_input = f.parameters[0]
f2 = contract.get_function_from_signature("f2(uint256,uint256)")

print(
    f"{var_dependant} is dependent of {user_input}: {is_dependent(var_dependant, user_input, contract)} (base)"
)
assert is_dependent(var_dependant, user_input, contract)
print(f"{var_tainted} is tainted: {is_tainted(var_tainted, contract)}")
assert is_tainted(var_tainted, contract)
print(f"{var_not_tainted} is tainted: {is_tainted(var_not_tainted, contract)}")
assert not is_tainted(var_not_tainted, contract)

print("SimpleModifier contract")
contracts = slither.get_contract_from_name("SimpleModifier")
assert len(contracts) == 1
contract = contracts[0]
dependent_state_var = contract.get_state_variable_from_name("owner")
assert dependent_state_var
baz = contract.get_modifier_from_signature("baz()")
intermediate = baz.get_local_variable_from_name("intermediate")
assert intermediate
print(
    f"{intermediate} depends on msg.sender: {is_dependent(intermediate, SolidityVariableComposed('msg.sender'), baz)}"
)
assert is_dependent(intermediate, SolidityVariableComposed("msg.sender"), baz)



"""
Module detecting constant functions
Recursively check the called functions
"""

from slither.core.compilation_unit import SlitherCompilationUnit
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    ALL_SOLC_VERSIONS_04,
    DETECTOR_INFO,
)
from slither.formatters.attributes.const_functions import custom_format
from slither.utils.output import Output


class ConstantFunctionsState(AbstractDetector):
    """
    Constant function detector
    """

    ARGUMENT = "constant-function-state"  # run the detector with slither.py --ARGUMENT
    HELP = "Constant functions changing the state"  # help information
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#constant-functions-changing-the-state"

    WIKI_TITLE = "Constant functions changing the state"

    # region wiki_description
    WIKI_DESCRIPTION = """
Functions declared as `constant`/`pure`/`view` change the state.

`constant`/`pure`/`view` was not enforced prior to Solidity 0.5.
Starting from Solidity 0.5, a call to a `constant`/`pure`/`view` function uses the `STATICCALL` opcode, which reverts in case of state modification.

As a result, a call to an [incorrectly labeled function may trap a contract compiled with Solidity 0.5](https://solidity.readthedocs.io/en/develop/050-breaking-changes.html#interoperability-with-older-contracts)."""
    # endregion wiki_description

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Constant{
    uint counter;
    function get() public view returns(uint){
       counter = counter +1;
       return counter
    }
}
```
`Constant` was deployed with Solidity 0.4.25. Bob writes a smart contract that interacts with `Constant` in Solidity 0.5.0.
All the calls to `get` revert, breaking Bob's smart contract execution."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = (
        "Ensure that attributes of contracts compiled prior to Solidity 0.5.0 are correct."
    )

    VULNERABLE_SOLC_VERSIONS = ALL_SOLC_VERSIONS_04

    def _detect(self) -> list[Output]:
        """Detect the constant function changing the state

        Recursively visit the calls
        Returns:
            list: {'vuln', 'filename,'contract','func','#varsWritten'}
        """
        results = []
        for c in self.contracts:
            for f in c.functions:
                if f.contract_declarer != c:
                    continue
                if f.view or f.pure:
                    variables_written = f.all_state_variables_written()
                    if variables_written:
                        attr = "view" if f.view else "pure"

                        info: DETECTOR_INFO = [
                            f,
                            f" is declared {attr} but changes state variables:\n",
                        ]

                        for variable_written in variables_written:
                            info += ["\t- ", variable_written, "\n"]

                        res = self.generate_result(info, {"contains_assembly": False})

                        results.append(res)

        return results

    @staticmethod
    def _format(slither: SlitherCompilationUnit, result: dict) -> None:
        custom_format(slither, result)



from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.output import Output
from slither.slithir.operations import Binary, Assignment, Unpack, SolidityCall
from slither.core.variables import Variable
from slither.core.declarations.solidity_variables import SolidityFunction
from slither.core.cfg.node import Node


class ChronicleUncheckedPrice(AbstractDetector):
    """
    Documentation: This detector finds calls to Chronicle oracle where the returned price is not checked
    https://docs.chroniclelabs.org/Resources/FAQ/Oracles#how-do-i-check-if-an-oracle-becomes-inactive-gets-deprecated
    """

    ARGUMENT = "chronicle-unchecked-price"
    HELP = "Detect when Chronicle price is not checked."
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#chronicle-unchecked-price"

    WIKI_TITLE = "Chronicle unchecked price"
    WIKI_DESCRIPTION = "Chronicle oracle is used and the price returned is not checked to be valid. For more information https://docs.chroniclelabs.org/Resources/FAQ/Oracles#how-do-i-check-if-an-oracle-becomes-inactive-gets-deprecated."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract C {
    IChronicle chronicle;

    constructor(address a) {
        chronicle = IChronicle(a);
    }

    function bad() public {
        uint256 price = chronicle.read();
    }
```
The `bad` function gets the price from Chronicle by calling the read function however it does not check if the price is valid."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Validate that the price returned by the oracle is valid."

    def _var_is_checked(self, nodes: list[Node], var_to_check: Variable) -> bool:
        visited = set()
        checked = False

        while nodes:
            if checked:
                break
            next_node = nodes[0]
            nodes = nodes[1:]

            for node_ir in next_node.all_slithir_operations():
                if isinstance(node_ir, Binary) and var_to_check in node_ir.read:
                    checked = True
                    break
                # This case is for tryRead and tryReadWithAge
                # if the isValid boolean is checked inside a require(isValid)
                if (
                    isinstance(node_ir, SolidityCall)
                    and node_ir.function
                    in (
                        SolidityFunction("require(bool)"),
                        SolidityFunction("require(bool,string)"),
                        SolidityFunction("require(bool,error)"),
                    )
                    and var_to_check in node_ir.read
                ):
                    checked = True
                    break

            if next_node not in visited:
                visited.add(next_node)
                for son in next_node.sons:
                    if son not in visited:
                        nodes.append(son)
        return checked

    def _detect(self) -> list[Output]:
        results: list[Output] = []

        for contract in self.compilation_unit.contracts_derived:
            for target_contract, ir in sorted(
                contract.all_high_level_calls,
                key=lambda x: (x[1].node.node_id, x[1].node.function.full_name),
            ):
                if target_contract.name in ("IScribe", "IChronicle") and ir.function_name in (
                    "read",
                    "tryRead",
                    "readWithAge",
                    "tryReadWithAge",
                    "latestAnswer",
                    "latestRoundData",
                ):
                    found = False
                    if ir.function_name in ("read", "latestAnswer"):
                        # We need to iterate the IRs as we are not always sure that the following IR is the assignment
                        # for example in case of type conversion it isn't
                        for node_ir in ir.node.irs:
                            if isinstance(node_ir, Assignment):
                                possible_unchecked_variable_ir = node_ir.lvalue
                                found = True
                                break
                    elif ir.function_name in ("readWithAge", "tryRead", "tryReadWithAge"):
                        # We are interested in the first item of the tuple
                        # readWithAge : value
                        # tryRead/tryReadWithAge : isValid
                        for node_ir in ir.node.irs:
                            if isinstance(node_ir, Unpack) and node_ir.index == 0:
                                possible_unchecked_variable_ir = node_ir.lvalue
                                found = True
                                break
                    elif ir.function_name == "latestRoundData":
                        found = False
                        for node_ir in ir.node.irs:
                            if isinstance(node_ir, Unpack) and node_ir.index == 1:
                                possible_unchecked_variable_ir = node_ir.lvalue
                                found = True
                                break

                    # If we did not find the variable assignment we know it's not checked
                    checked = (
                        self._var_is_checked(ir.node.sons, possible_unchecked_variable_ir)
                        if found
                        else False
                    )

                    if not checked:
                        info: DETECTOR_INFO = [
                            "Chronicle price is not checked to be valid in ",
                            ir.node.function,
                            "\n\t- ",
                            ir.node,
                            "\n",
                        ]
                        res = self.generate_result(info)
                        results.append(res)

        return results



from slither.core.solidity_types.type import Type
from slither.slithir.operations.lvalue import OperationWithLValue
from slither.slithir.variables.temporary import TemporaryVariable


class TmpNewArray(OperationWithLValue):
    def __init__(
        self,
        array_type: Type,
        lvalue: TemporaryVariable,
    ) -> None:
        super().__init__()
        assert isinstance(array_type, Type)
        self._array_type = array_type
        self._lvalue = lvalue

    @property
    def array_type(self) -> Type:
        return self._array_type

    @property
    def read(self):
        return []

    def __str__(self):
        return f"{self.lvalue} = new {self.array_type}"



"""
Module printing summary of the contract
"""

from slither.core.declarations import SolidityFunction
from slither.printers.abstract_printer import AbstractPrinter
from slither.slithir.operations import SolidityCall
from slither.utils.myprettytable import MyPrettyTable

require_or_assert = [
    SolidityFunction("assert(bool)"),
    SolidityFunction("require(bool)"),
    SolidityFunction("require(bool,string)"),
    SolidityFunction("require(bool,error)"),
]


class RequireOrAssert(AbstractPrinter):
    ARGUMENT = "require"
    HELP = "Print the require and assert calls of each function"

    WIKI = "https://github.com/trailofbits/slither/wiki/Printer-documentation#require"

    @staticmethod
    def _convert(l):
        return "\n".join(l)

    def output(self, _filename):
        """
        _filename is not used
        Args:
            _filename(string)
        """

        all_tables = []
        all_txt = ""
        for contract in self.slither.contracts_derived:
            txt = f"\nContract {contract.name}"
            table = MyPrettyTable(["Function", "require or assert"])
            for function in contract.functions:
                require = function.all_slithir_operations()
                require = [
                    ir
                    for ir in require
                    if isinstance(ir, SolidityCall) and ir.function in require_or_assert
                ]
                require = [ir.node for ir in require]
                table.add_row(
                    [
                        function.name,
                        self._convert(sorted([str(m.expression) for m in set(require)])),
                    ]
                )
            txt += "\n" + str(table)
            self.info(txt)
            all_tables.append((contract.name, table))
            all_txt += txt

        res = self.generate_output(all_txt)
        for name, table in all_tables:
            res.add_pretty_table(table, name)

        return res



import sys
from slither.slither import Slither

if len(sys.argv) != 2:
    print("python variable_in_condition.py variable_in_condition.sol")
    sys.exit(-1)

# Init slither
slither = Slither(sys.argv[1])

# Get the contract
contracts = slither.get_contract_from_name("Contract")
assert len(contracts) == 1
contract = contracts[0]
# Get the variable
var_a = contract.get_state_variable_from_name("a")
assert var_a

# Get the functions reading the variable
functions_reading_a = contract.get_functions_reading_from_variable(var_a)

function_using_a_as_condition = [
    f
    for f in functions_reading_a
    if f.is_reading_in_conditional_node(var_a) or f.is_reading_in_require_or_assert(var_a)
]

# Print the result
print(f'The function using "a" in condition are {[f.name for f in function_using_a_as_condition]}')



from .assignment_operation import AssignmentOperation, AssignmentOperationType
from .binary_operation import BinaryOperation, BinaryOperationType
from .call_expression import CallExpression
from .conditional_expression import ConditionalExpression
from .elementary_type_name_expression import ElementaryTypeNameExpression
from .identifier import Identifier
from .index_access import IndexAccess
from .literal import Literal
from .member_access import MemberAccess
from .new_array import NewArray
from .new_contract import NewContract
from .new_elementary_type import NewElementaryType
from .super_call_expression import SuperCallExpression
from .super_identifier import SuperIdentifier
from .self_identifier import SelfIdentifier
from .tuple_expression import TupleExpression
from .type_conversion import TypeConversion
from .unary_operation import UnaryOperation, UnaryOperationType
